"""Abstract finite element classes and functions."""

from __future__ import annotations

import math
import typing
from abc import ABC, abstractmethod
from itertools import product

import sympy

from .basis_functions import BasisFunction
from .functionals import ListOfFunctionals
from .functions import (AnyFunction, FunctionInput, ScalarFunction, VectorFunction,
                        parse_function_input)
from .geometry import PointType, SetOfPointsInput, parse_set_of_points_input
from .piecewise_functions import PiecewiseFunction
from .plotting import Picture, colors
from .references import Reference
from .symbols import x
from .utils import allequal

TabulatedBasis = typing.Union[
    typing.List[typing.Union[sympy.core.expr.Expr, int]],
    typing.List[typing.Tuple[typing.Union[sympy.core.expr.Expr, int], ...]],
    typing.List[typing.Tuple[typing.Tuple[typing.Union[sympy.core.expr.Expr, int], ...], ...]],
    typing.List[sympy.matrices.dense.MutableDenseMatrix],
]


class NoTensorProduct(Exception):
    """Error for element without a tensor representation."""

    def __init__(self):
        """Initialise."""
        super().__init__("This element does not have a tensor product representation.")


class FiniteElement(ABC):
    """Abstract finite element."""

    _float_basis_functions: typing.Union[None, typing.List[AnyFunction]]
    _value_scale: typing.Union[None, sympy.core.expr.Expr]

    def __init__(
        self, reference: Reference, order: int, space_dim: int, domain_dim: int, range_dim: int,
        range_shape: typing.Optional[typing.Tuple[int, ...]] = None
    ):
        """Create a finite element.

        Args:
            reference: The reference cell
            order: The polynomial order
            space_dim: The dimension of the finite element space
            domain_dim: The topological dimension of the reference cell
            range_dim: The dimension/value size of functions in the space
            range_shape: The (value) shape of function in the space
        """
        self.reference = reference
        self.order = order
        self.space_dim = space_dim
        self.domain_dim = domain_dim
        self.range_dim = range_dim
        self.range_shape = range_shape
        self._float_basis_functions = None
        self._value_scale = None

    @abstractmethod
    def plot_dof_diagram(
        self, filename: typing.Union[str, typing.List[str]],
        plot_options: typing.Dict[str, typing.Any] = {}, **kwargs: typing.Any
    ):
        """Plot a diagram showing the DOFs of the element.

        Args:
            filename: The file name
            plot_options: Options for the plot
            kwargs: Keyword arguments
        """
        pass

    @abstractmethod
    def entity_dofs(self, entity_dim: int, entity_number: int) -> typing.List[int]:
        """Get the numbers of the DOFs associated with the given entity.

        Args:
            entity_dim: The dimension of the entity
            entity_number: The number of the entity

        Returns:
            The numbers of the DOFs associated with the entity
        """
        pass

    @abstractmethod
    def get_basis_functions(
        self, use_tensor_factorisation: bool = False
    ) -> typing.List[AnyFunction]:
        """Get the basis functions of the element.

        Args:
            use_tensor_factorisation: Should a tensor factorisation be used?

        Returns:
            The basis functions
        """
        pass

    def get_basis_function(self, n: int) -> BasisFunction:
        """Get a single basis function of the element.

        Args:
            n: The number of the basis function

        Returns:
            The basis function
        """
        return ElementBasisFunction(self, n)

    def tabulate_basis(
        self, points_in: SetOfPointsInput, order: str = "xyzxyz",
    ) -> TabulatedBasis:
        """Evaluate the basis functions of the element at the given points.

        Args:
            points_in: The points
            order: The order to return the values

        Returns:
            The tabulated basis functions
        """
        points = parse_set_of_points_input(points_in)
        tabbed = [tuple(b.subs(x, p).as_sympy() for b in self.get_basis_functions())
                  for p in points]
        if self.range_dim == 1:
            return tabbed

        if order == "xyz,xyz":
            return tabbed
        elif order == "xxyyzz":
            output = []
            for row in tabbed:
                output.append(tuple(j for i in zip(*row) for j in i))
            return output
        elif order == "xyzxyz":
            output = []
            for row in tabbed:
                output.append(tuple(j for i in row for j in i))
            return output
        else:
            raise ValueError(f"Unknown order: {order}")

    def tabulate_basis_float(self, points_in: SetOfPointsInput) -> TabulatedBasis:
        """Evaluate the basis functions of the element at the given points in xyz,xyz order.

        Args:
            points_in: The points

        Returns:
            The tabulated basis functions
        """
        if self._float_basis_functions is None:
            self._float_basis_functions = [b.with_floats() for b in self.get_basis_functions()]

        assert self._float_basis_functions is not None
        points = parse_set_of_points_input(points_in)
        return [tuple(b.subs(x, p).as_sympy() for b in self._float_basis_functions) for p in points]

    def plot_basis_function(
        self, n: int, filename: typing.Union[str, typing.List[str]], **kwargs: typing.Any
    ):
        """Plot a diagram showing a basis function.

        Args:
            n: The basis function number
            filename: The file name
            kwargs: Keyword arguments
        """
        if self._value_scale is None:
            max_v = 0.0
            values = self.tabulate_basis_float(self.reference.make_lattice_float(6))
            for row in values:
                assert isinstance(row, tuple)
                for i in row:
                    max_v = max(max_v, float(parse_function_input(i).norm()))
            self._value_scale = 1 / sympy.Float(max_v)

        f = self.get_basis_functions()[n]
        assert self._value_scale is not None
        f.plot(self.reference, filename, None, None, None, n, self._value_scale, **kwargs)

    @abstractmethod
    def map_to_cell(
        self, vertices_in: SetOfPointsInput,
        basis: typing.Optional[typing.List[AnyFunction]] = None,
        forward_map: typing.Optional[PointType] = None,
        inverse_map: typing.Optional[PointType] = None
    ) -> typing.List[AnyFunction]:
        """Map the basis onto a cell using the appropriate mapping for the element.

        Args:
            vertices_in: The vertices of the cell
            basis: The basis functions
            forward_map: The map from the reference to the cell
            inverse_map: The map to the reference from the cell

        Returns:
            The basis functions mapped to the cell
        """
        pass

    @abstractmethod
    def get_polynomial_basis(self) -> typing.List[AnyFunction]:
        """Get the symbolic polynomial basis for the element.

        Returns:
            The polynomial basis
        """
        pass

    def test(self):
        """Run tests for this element."""
        if self.order <= 4:
            self.test_continuity()

    def test_continuity(self):
        """Test that this element has the correct continuity."""
        continuity = self.continuity
        if "{order}" in continuity:
            continuity = continuity.replace("{order}", f"{self.order}")

        # Test continuity
        if self.reference.name == "interval":
            vertices = ((-1, ), (0, ))
            entity_pairs = [[0, (0, 1)]]
        elif self.reference.name == "triangle":
            vertices = ((-1, 0), (0, 0), (0, 1))
            entity_pairs = [[0, (0, 1)], [0, (2, 2)], [1, (1, 0)]]
        elif self.reference.name == "tetrahedron":
            vertices = ((-1, 0, 0), (0, 0, 0), (0, 1, 0), (0, 0, 1))
            entity_pairs = [[0, (0, 1)], [0, (2, 2)], [0, (3, 3)],
                            [1, (0, 0)], [1, (3, 1)], [1, (4, 2)],
                            [2, (1, 0)]]
        elif self.reference.name == "quadrilateral":
            vertices = ((0, 0), (0, 1), (-1, 0), (-1, 1))
            entity_pairs = [[0, (0, 0)], [0, (2, 1)], [1, (1, 0)]]
        elif self.reference.name == "hexahedron":
            vertices = ((0, 0, 0), (0, 0, 1), (0, 1, 0), (0, 1, 1),
                        (-1, 0, 0), (-1, 0, 1), (-1, 1, 0), (-1, 1, 1))
            entity_pairs = [[0, (0, 0)], [0, (2, 2)], [0, (4, 1)], [0, (6, 3)],
                            [1, (1, 1)], [1, (2, 0)], [1, (6, 5)], [1, (9, 3)],
                            [2, (0, 2)]]
        elif self.reference.name == "prism":
            vertices = ((-1, 0, 0), (0, 0, 0), (0, 1, 0),
                        (-1, 0, 1), (0, 0, 1), (0, 1, 1))
            entity_pairs = [[0, (0, 1)], [0, (2, 2)], [0, (3, 4)], [0, (5, 5)],
                            [1, (1, 3)], [1, (2, 4)], [1, (6, 6)], [1, (7, 8)],
                            [2, (2, 3)]]
        elif self.reference.name == "pyramid":
            vertices = ((-1, 0, 0), (0, 0, 0), (-1, 1, 0),
                        (0, 1, 0), (0, 0, 1))
            entity_pairs = [[0, (0, 1)], [0, (2, 3)], [0, (4, 4)],
                            [1, (1, 3)], [1, (2, 4)], [1, (6, 7)],
                            [2, (2, 3)]]

        if continuity == "L2":
            return

        for dim, entities in entity_pairs:
            for fi, gi in zip(*[self.entity_dofs(dim, i) for i in entities]):
                basis = self.get_basis_functions()
                try:
                    basis2 = self.map_to_cell(vertices)
                except NotImplementedError:
                    return "Mapping not implemented for this element."

                f = basis[fi]
                g = basis2[gi]

                def get_piece(f, point):
                    if isinstance(f, PiecewiseFunction):
                        return f.get_piece(point)
                    if isinstance(f, list):
                        return [get_piece(g, point) for g in f]
                    if isinstance(f, tuple):
                        return tuple(get_piece(g, point) for g in f)
                    return f

                if self.reference.tdim == 2:
                    f = get_piece(f, (0, sympy.Rational(1, 2)))
                    g = get_piece(g, (0, sympy.Rational(1, 2)))
                elif self.reference.tdim == 3:
                    f = get_piece(f, (0, sympy.Rational(1, 3), sympy.Rational(1, 3)))
                    g = get_piece(g, (0, sympy.Rational(1, 3), sympy.Rational(1, 3)))

                f = f.subs(x[0], 0)
                g = g.subs(x[0], 0)

                if continuity[0] == "C":
                    order = int(continuity[1:])
                    deriv_f = [f]
                    deriv_g = [g]
                    f = [f]
                    g = [g]
                    for _ in range(order):
                        deriv_f = [d.diff(i) for d in deriv_f for i in x[:self.reference.tdim]]
                        f += deriv_f
                        deriv_g = [d.diff(i) for d in deriv_g for i in x[:self.reference.tdim]]
                        g += deriv_g
                elif continuity == "H(div)":
                    f = f[0]
                    g = g[0]
                elif continuity == "inner H(div)":
                    f = f[0, 0]
                    g = g[0, 0]
                elif continuity == "H(curl)":
                    f = f[1:]
                    g = g[1:]
                elif continuity == "inner H(curl)":
                    if len(vertices[0]) == 2:
                        f = f[1, 1]
                        g = g[1, 1]
                    if len(vertices[0]) == 3:
                        if dim == 1:
                            vs = self.reference.sub_entities(1)[entities[0]]
                            v0 = self.reference.vertices[vs[0]]
                            v1 = self.reference.vertices[vs[1]]
                            tangent = VectorFunction(v1) - VectorFunction(v0)
                            f = sum(i * f[ni, nj] * j
                                    for ni, i in enumerate(tangent)
                                    for nj, j in enumerate(tangent))
                            g = sum(i * g[ni, nj] * j
                                    for ni, i in enumerate(tangent)
                                    for nj, j in enumerate(tangent))
                        else:
                            assert dim == 2
                            f = [f[1, 1], f[2, 2]]
                            g = [g[1, 1], g[2, 2]]
                elif continuity == "integral inner H(div)":
                    f = f[0, 0].integrate((x[1], 0, 1))
                    g = g[0, 0].integrate((x[1], 0, 1))
                else:
                    raise ValueError(f"Unknown continuity: {continuity}")

                assert allequal(f, g)

    def get_tensor_factorisation(
        self
    ) -> typing.List[typing.Tuple[str, typing.List[FiniteElement], typing.List[int]]]:
        """Get the representation of the element as a tensor product.

        Returns:
            The tensor factorisation
        """
        raise NoTensorProduct()

    def _get_basis_functions_tensor(self) -> typing.List[AnyFunction]:
        """Compute the basis functions using the space's tensor product factorisation.

        Returns:
            The basis functions
        """
        factorisation = self.get_tensor_factorisation()
        basis = {}
        for t_type, factors, perm in factorisation:
            if t_type == "scalar":
                tensor_bases = [[i.subs(x[0], x_i) for i in f.get_basis_functions()]
                                for x_i, f in zip(x, factors)]
                for p, k in zip(perm, product(*tensor_bases)):
                    basis[p] = ScalarFunction(1)
                    for i in k:
                        basis[p] *= i
            else:
                raise ValueError(f"Unknown tensor product type: {t_type}")
        return [basis[i] for i in range(len(basis))]

    @property
    def name(self) -> str:
        """Get the name of the element.

        Returns:
            The name of the element's family
        """
        return self.names[0]

    names: typing.List[str] = []
    references: typing.List[str] = []


class CiarletElement(FiniteElement):
    """Finite element defined using the Ciarlet definition."""

    def __init__(
        self, reference: Reference, order: int, basis: typing.List[FunctionInput],
        dofs: ListOfFunctionals, domain_dim: int, range_dim: int,
        range_shape: typing.Optional[typing.Tuple[int, ...]] = None
    ):
        """Create a Ciarlet element.

        Args:
            reference: The reference cell
            order: The polynomial order
            basis: The polynomial basis
            dofs: The DOF functionals
            domain_dim: The topological dimension of the cell
            range_dim: The dimension of the range
            range_shape: The shape of the range
        """
        super().__init__(reference, order, len(dofs), domain_dim, range_dim, range_shape)
        assert len(basis) == len(dofs)
        self._basis = [parse_function_input(b) for b in basis]
        for b in self._basis:
            assert isinstance(b, AnyFunction)
        self.dofs = dofs
        self._basis_functions: typing.Union[typing.List[AnyFunction], None] = None

    def entity_dofs(self, entity_dim: int, entity_number: int) -> typing.List[int]:
        """Get the numbers of the DOFs associated with the given entity.

        Args:
            entity_dim: The dimension of the entity
            entity_number: The number of the entity

        Returns:
            The numbers of the DOFs associated with the entity
        """
        return [i for i, j in enumerate(self.dofs) if j.entity == (entity_dim, entity_number)]

    def get_polynomial_basis(self) -> typing.List[AnyFunction]:
        """Get the symbolic polynomial basis for the element.

        Returns:
            The polynomial basis
        """
        return self._basis

    def get_dual_matrix(self) -> sympy.matrices.dense.MutableDenseMatrix:
        """Get the dual matrix.

        Returns:
            The dual matrix
        """
        mat = []
        for b in self.get_polynomial_basis():
            row = []
            for d in self.dofs:
                entry = d.eval_symbolic(b).as_sympy()
                row.append(entry)
            mat.append(row)
        return sympy.Matrix(mat)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the keyword arguments used to create this element.

        Returns:
            Keyword arguments dictionary
        """
        return {}

    def get_basis_functions(
        self, use_tensor_factorisation: bool = False
    ) -> typing.List[AnyFunction]:
        """Get the basis functions of the element.

        Args:
            use_tensor_factorisation: Should a tensor factorisation be used?

        Returns:
            The basis functions
        """
        if self._basis_functions is None:
            if use_tensor_factorisation:
                self._basis_functions = self._get_basis_functions_tensor()
            else:
                m = self.get_dual_matrix()
                assert isinstance(m, sympy.Matrix)
                minv = m.inv("LU")

                sfs: typing.List[AnyFunction] = []
                pb = self.get_polynomial_basis()
                for i, dof in enumerate(self.dofs):
                    sf = ScalarFunction(minv[i, 0]) * pb[0]
                    for c, d in zip(minv.row(i)[1:], pb[1:]):
                        sf += ScalarFunction(c) * d
                    sfs.append(sf)

                self._basis_functions = sfs

        assert isinstance(self._basis_functions, list)

        return self._basis_functions

    def plot_basis_function(
        self, n: int, filename: typing.Union[str, typing.List[str]], **kwargs: typing.Any
    ):
        """Plot a diagram showing a basis function.

        Args:
            n: The basis function number
            filename: The file name
            kwargs: Keyword arguments
        """
        if self._value_scale is None:
            values = self.tabulate_basis_float(self.reference.make_lattice_float(6))
            max_v = 0.0
            for row in values:
                assert isinstance(row, tuple)
                for i in row:
                    max_v = max(max_v, float(parse_function_input(i).norm()))
            self._value_scale = 1 / sympy.Float(max_v)

        f = self.get_basis_functions()[n]
        d = self.dofs[n]
        assert self._value_scale is not None
        f.plot(self.reference, filename, d.dof_point(), d.dof_direction(), d.entity, n,
               self._value_scale, **kwargs)

    def plot_dof_diagram(
        self, filename: typing.Union[str, typing.List[str]],
        plot_options: typing.Dict[str, typing.Any] = {}, **kwargs: typing.Any
    ):
        """Plot a diagram showing the DOFs of the element.

        Args:
            filename: The file name
            plot_options: Options for the plot
            kwargs: Keyword arguments
        """
        img = Picture(**kwargs)

        dofs_by_subentity: typing.Dict[int, typing.Dict[int, ListOfFunctionals]] = {
            i: {j: [] for j in range(self.reference.sub_entity_count(i))}
            for i in range(self.reference.tdim + 1)}

        for d in self.dofs:
            dofs_by_subentity[d.entity[0]][d.entity[1]].append(d)

        for entities in self.reference.z_ordered_entities():
            for dim, e in entities:
                if dim == 1:
                    pts = tuple(self.reference.vertices[i] for i in self.reference.edges[e])
                    img.add_line(pts[0], pts[1], colors.BLACK)
                if dim == 2:
                    pts = tuple(self.reference.vertices[i] for i in self.reference.faces[e])
                    if len(pts) == 4:
                        pts = (pts[0], pts[1], pts[3], pts[2])
                    img.add_fill(pts, colors.WHITE, 0.5)

            for dim, e in entities:
                dofs = dofs_by_subentity[dim][e]
                dofs.sort(key=lambda d: img.z(d.adjusted_dof_point()))
                for d in dofs:
                    direction = d.dof_direction()
                    if direction is not None:
                        shifted = False
                        for d2 in self.dofs:
                            if d != d2 and d.adjusted_dof_point() == d2.adjusted_dof_point():
                                shifted = True
                                break
                        img.add_dof_arrow(d.adjusted_dof_point(), direction, self.dofs.index(d),
                                          colors.entity(d.entity[0]), shifted)
                    else:
                        img.add_dof_marker(
                            d.adjusted_dof_point(), self.dofs.index(d), colors.entity(d.entity[0]))

        img.save(filename, plot_options=plot_options)

    def map_to_cell(
        self, vertices_in: SetOfPointsInput,
        basis: typing.Optional[typing.List[AnyFunction]] = None,
        forward_map: typing.Optional[PointType] = None,
        inverse_map: typing.Optional[PointType] = None
    ) -> typing.List[AnyFunction]:
        """Map the basis onto a cell using the appropriate mapping for the element.

        Args:
            vertices_in: The vertices of the cell
            basis: The basis functions
            forward_map: The map from the reference to the cell
            inverse_map: The map to the reference from the cell

        Returns:
            The basis functions mapped to the cell
        """
        vertices = parse_set_of_points_input(vertices_in)
        if basis is None:
            basis = self.get_basis_functions()
        if forward_map is None:
            forward_map = self.reference.get_map_to(vertices)
        if inverse_map is None:
            inverse_map = self.reference.get_inverse_map_to(vertices)

        functions: typing.List[AnyFunction] = [ScalarFunction(0) for f in basis]
        for dim in range(self.reference.tdim + 1):
            for e in range(self.reference.sub_entity_count(dim)):
                entity_dofs = self.entity_dofs(dim, e)
                dofs_by_type: typing.Dict[
                    typing.Tuple[typing.Type, typing.Union[str, None]], typing.List[int]
                ] = {}
                for d in entity_dofs:
                    dof = self.dofs[d]
                    t = (type(dof), dof.mapping)
                    if t not in dofs_by_type:
                        dofs_by_type[t] = []
                    dofs_by_type[t].append(d)
                for ds in dofs_by_type.values():
                    mapped_dofs = self.dofs[ds[0]].perform_mapping(
                        [basis[d] for d in ds],
                        forward_map, inverse_map, self.reference.tdim)
                    for d_n, mdof in zip(ds, mapped_dofs):
                        functions[d_n] = mdof

        for fun in functions:
            if isinstance(fun, PiecewiseFunction):
                fun.map_pieces(forward_map)

        return functions

    def test(self):
        """Run tests for this element."""
        super().test()
        self.test_functional_entities()
        self.test_functionals()
        self.test_dof_points()

    def test_dof_points(self):
        """Test that DOF points are valid."""
        for d in self.dofs:
            p = d.dof_point()
            assert len(p) == self.reference.gdim
            for i in p:
                if i is None:
                    break
                else:
                    assert not math.isnan(float(i))
            else:
                assert self.reference.contains(p)

    def test_functional_entities(self):
        """Test that the dof entities are valid and match the references of integrals."""
        for dof in self.dofs:
            dim, entity = dof.entity
            assert entity < self.reference.sub_entity_count(dim)
            if hasattr(dof, "integral_domain"):
                assert dim == dof.integral_domain.tdim

    def test_functionals(self):
        """Test that the functionals are satisfied by the basis functions."""
        for i, f in enumerate(self.get_basis_functions()):
            for j, d in enumerate(self.dofs):
                if i == j:
                    assert d.eval(f) == 1
                else:
                    assert d.eval(f) == 0
                assert d.entity_dim() is not None


class DirectElement(FiniteElement):
    """Finite element defined directly."""

    _basis_functions: typing.List[AnyFunction]

    def __init__(
        self, reference: Reference, order: int, basis_functions: typing.List[FunctionInput],
        basis_entities: typing.List[typing.Tuple[int, int]],
        domain_dim: int, range_dim: int, range_shape: typing.Optional[typing.Tuple[int, ...]] = None
    ):
        """Create a Ciarlet element.

        Args:
            reference: The reference cell
            order: The polynomial order
            basis_functions: The basis functions
            basis_entities: The entitiy each basis function is associated with
            domain_dim: The topological dimension of the cell
            range_dim: The dimension of the range
            range_shape: The shape of the range
        """
        super().__init__(reference, order, len(basis_functions), domain_dim, range_dim,
                         range_shape)
        self._basis_entities = basis_entities
        self._basis_functions = [parse_function_input(f) for f in basis_functions]

    def entity_dofs(self, entity_dim: int, entity_number: int) -> typing.List[int]:
        """Get the numbers of the DOFs associated with the given entity.

        Args:
            entity_dim: The dimension of the entity
            entity_number: The number of the entity

        Returns:
            The numbers of the DOFs associated with the entity
        """
        return [i for i, j in enumerate(self._basis_entities) if j == (entity_dim, entity_number)]

    def get_basis_functions(
        self, use_tensor_factorisation: bool = False
    ) -> typing.List[AnyFunction]:
        """Get the basis functions of the element.

        Args:
            use_tensor_factorisation: Should a tensor factorisation be used?

        Returns:
            The basis functions
        """
        if use_tensor_factorisation:
            return self._get_basis_functions_tensor()

        return self._basis_functions

    def map_to_cell(
        self, vertices_in: SetOfPointsInput,
        basis: typing.Optional[typing.List[AnyFunction]] = None,
        forward_map: typing.Optional[PointType] = None,
        inverse_map: typing.Optional[PointType] = None
    ) -> typing.List[AnyFunction]:
        """Map the basis onto a cell using the appropriate mapping for the element.

        Args:
            vertices_in: The vertices of the cell
            basis: The basis functions
            forward_map: The map from the reference to the cell
            inverse_map: The map to the reference from the cell

        Returns:
            The basis functions mapped to the cell
        """
        raise NotImplementedError()

    def get_polynomial_basis(self) -> typing.List[AnyFunction]:
        """Get the symbolic polynomial basis for the element.

        Returns:
            The polynomial basis
        """
        raise NotImplementedError()

    def plot_dof_diagram(
        self, filename: typing.Union[str, typing.List[str]],
        plot_options: typing.Dict[str, typing.Any] = {}, **kwargs: typing.Any
    ):
        """Plot a diagram showing the DOFs of the element.

        Args:
            filename: The file name
            plot_options: Options for the plot
            kwargs: Keyword arguments
        """
        img = Picture(**kwargs)

        dofs_by_subentity: typing.Dict[int, typing.Dict[int, typing.List[int]]] = {
            i: {j: [] for j in range(self.reference.sub_entity_count(i))}
            for i in range(self.reference.tdim + 1)}

        for i, e in enumerate(self._basis_entities):
            dofs_by_subentity[e[0]][e[1]].append(i)

        for entities in self.reference.z_ordered_entities():
            for dim, e_n in entities:
                if dim == 1:
                    pts = tuple(self.reference.vertices[i] for i in self.reference.edges[e_n])
                    img.add_line(pts[0], pts[1], colors.BLACK)
                if dim == 2:
                    pts = tuple(self.reference.vertices[i] for i in self.reference.faces[e_n])
                    if len(pts) == 4:
                        pts = (pts[0], pts[1], pts[3], pts[2])
                    img.add_fill(pts, colors.WHITE, 0.5)

            for dim, e_n in entities:
                n = len(dofs_by_subentity[dim][e_n])
                if n > 0:
                    sub_ref = self.reference.sub_entity(dim, e_n)
                    points: typing.List[PointType] = []
                    if dim == 0:
                        assert n == 1
                        points = [sub_ref.vertices[0]]
                    elif dim == 1:
                        points = [tuple(o + sympy.Rational(i * a, n + 1)
                                        for o, a in zip(sub_ref.origin, *sub_ref.axes))
                                  for i in range(1, n + 1)]
                    elif dim == 2:
                        ne = 1
                        while ne * (ne + 1) // 2 < n:
                            ne += 1
                        points = [tuple(o + sympy.Rational(i * a + j * b, n + 1)
                                        for o, a, b in zip(sub_ref.origin, *sub_ref.axes))
                                  for i in range(1, ne + 1) for j in range(1, ne + 1 - i)]
                    elif dim == 3:
                        ne = 1
                        while ne * (ne + 1) * (ne + 2) // 6 < n:
                            ne += 1
                        points = [tuple(o + sympy.Rational(i * a + j * b + k * c, n + 1)
                                        for o, a, b, c in zip(sub_ref.origin, *sub_ref.axes))
                                  for i in range(1, ne + 1) for j in range(1, ne + 1 - i)
                                  for k in range(1, ne + 1 - i - j)]

                    for p, d in zip(points, dofs_by_subentity[dim][e_n]):
                        img.add_dof_marker(p, d, colors.entity(dim))

        img.save(filename, plot_options=plot_options)

    def test(self):
        """Run tests for this element."""
        super().test()
        self.test_independence()

    def test_independence(self):
        """Test that the basis functions of this element are linearly independent."""
        basis = self.get_basis_functions()
        all_terms = set()

        scalar = basis[0].is_scalar

        if scalar:
            for f in basis:
                f_s = f.as_sympy()
                assert isinstance(f_s, sympy.core.expr.Expr)
                for term in f_s.as_coefficients_dict():
                    all_terms.add(term)
            mat = [[0 for i in all_terms] for j in basis]
            for i, t in enumerate(all_terms):
                for j, f in enumerate(basis):
                    f_s = f.as_sympy()
                    assert isinstance(f_s, sympy.core.expr.Expr)
                    fd = f_s.as_coefficients_dict()
                    if t in fd:
                        mat[j][i] = fd[t]
        else:
            for f in basis:
                for fi, fpart in enumerate(f):
                    fpart_s = fpart.as_sympy()
                    assert isinstance(fpart_s, sympy.core.expr.Expr)
                    for term in fpart_s.as_coefficients_dict():
                        all_terms.add((fi, term))
            mat = [[0 for i in all_terms] for j in basis]
            for i, (fi, t) in enumerate(all_terms):
                for j, f in enumerate(basis):
                    ffi_s = f[fi].as_sympy()
                    assert isinstance(ffi_s, sympy.core.expr.Expr)
                    fd = ffi_s.as_coefficients_dict()
                    if t in fd:
                        mat[j][i] = fd[t]
        mat = sympy.Matrix(mat)

        assert mat.rank() == mat.rows


class ElementBasisFunction(BasisFunction):
    """A basis function of a finite element."""

    def __init__(self, element: FiniteElement, n: int):
        """Create an element basis function.

        Args:
            element: The finite element
            n: The basis function number
        """
        if element.range_dim == 1:
            super().__init__(scalar=True)
        else:
            if element.range_shape is None or len(element.range_shape) == 1:
                super().__init__(vector=True)
            else:
                assert len(element.range_shape) == 2
                super().__init__(matrix=True)
        self.element = element
        self.n = n

    def get_function(self) -> AnyFunction:
        """Get the actual basis function.

        Returns:
            The basis function
        """
        return self.element.get_basis_functions()[self.n]



"""Test P1-iso-P2 elements."""

import pytest
import sympy

import symfem
from symfem.symbols import x


@pytest.mark.parametrize("cell", ["triangle", "quadrilateral"])
def test_polyset(cell):
    e = symfem.create_element(cell, "P1-iso-P2", 1)

    half = sympy.Rational(1, 2)
    if cell == "triangle":
        points = [(0, 0), (1, 0), (0, 1), (half, 0), (0, half), (half, half)]
    elif cell == "quadrilateral":
        points = [(0, 0), (1, 0), (0, 1), (half, 0), (0, half), (half, half)]
    else:
        raise ValueError(f"Unsupported cell: {cell}")

    for f in e.get_polynomial_basis():
        for p in points:
            value = None
            for piece in f.pieces:
                if p in piece[0]:
                    if value is None:
                        value = piece[1].subs(x, p)
                    assert piece[1].subs(x, p) == value



"""Test documentation."""

import json
import os
import re
import sys
import typing
from io import StringIO

import pytest

import symfem

code = False
output = False
check_for_output = False
lines: typing.List[str] = []
doc_data = []
outputlines: typing.List[str] = []
codelines: typing.List[str] = []

if os.path.isfile(os.path.join(os.path.dirname(os.path.realpath(__file__)),
                  "../docs/index.rst")):
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)),
              "../docs/index.rst")) as f:
        for line in f:
            if line.strip() == ".. code-block:: python":
                code = True
                output = False
                codelines = []
                outputlines = []
            elif code and line.strip() == "::":
                output = True
                code = False
            elif code or output:
                if line.strip() != "":
                    if line.startswith("    "):
                        if code:
                            assert not output
                            codelines.append(line.strip("\n")[4:])
                        else:
                            assert output
                            outputlines.append(line.strip("\n")[4:])
                    else:
                        output = False
                        code = False
                        doc_data.append(("\n".join(codelines), "\n".join(outputlines)))
if os.path.isfile(os.path.join(os.path.dirname(os.path.realpath(__file__)),
                  "../README.md")):
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)),
              "../README.md")) as f:
        for line in f:
            if line.strip() != "":
                if line.strip() == "```python":
                    check_for_output = False
                    code = True
                    output = False
                    codelines = []
                    outputlines = []
                elif line.strip() == "```":
                    if code:
                        code = False
                        check_for_output = True
                    elif output:
                        output = False
                    elif check_for_output:
                        check_for_output = False
                        output = True
                elif check_for_output:
                    check_for_output = False
                elif code or output:
                    if code:
                        assert not output
                        codelines.append(line.strip("\n"))
                    else:
                        assert output
                        outputlines.append(line.strip("\n"))
                if len(codelines) > 0 and not code and not output and not check_for_output:
                    doc_data.append(("\n".join(codelines), "\n".join(outputlines)))
                    codelines = []
                    outputlines = []


def test_available_references():
    r_supported = symfem.create_reference.__doc__
    r_supported = r_supported.split("cell_type: ")[1]
    r_supported = r_supported.split("vertices: ")[0]
    r_supported = r_supported.split("Supported values:")[1]
    r_supported = set([i.strip() for i in r_supported.split(",")])

    e_supported = symfem.create_element.__doc__
    e_supported = e_supported.split("cell_type: ")[1]
    e_supported = e_supported.split("element_type: ")[0]
    e_supported = e_supported.split("Supported values:")[1]
    e_supported = set([i.strip() for i in e_supported.split(",")])

    assert r_supported == e_supported


def test_available_elements():
    supported = symfem.create_element.__doc__
    supported = supported.split("element_type: ")[1]
    supported = supported.split("order: ")[0]
    supported = supported.split("Supported values:")[1]
    supported = set([i.strip() for i in supported.split(",")])

    assert set(symfem.create._elementmap.keys()) == supported


def test_readme_elements():
    elementlist = {}
    for e in symfem.create._elementlist:
        for r in e.references:
            if r not in elementlist:
                elementlist[r] = []
            elementlist[r].append(e.names[0])

    root = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")
    with open(os.path.join(root, "README.md")) as f:
        readme = f.read().split("# Available cells and elements")[1]
    cells = []
    for r in readme.split("\n## ")[1:]:
        lines = r.split("\n")
        cell = r.split("\n")[0].strip().lower()
        cells.append(cell)
        lines = r.split("### List of supported elements")[1].strip().split("\n")
        elements = [i[2:].split("(alternative names:")[0].strip()
                    for i in lines if i.strip() != ""]

        for line in r.split("\n"):
            if "(alternative names:" in line:
                n, others = line.split("(alternative names:")
                names = [n[1:].strip()] + [i.strip() for i in others.split(")")[0].split(",")]
                for e in symfem.create._elementlist:
                    if names[0] in e.names and cell in e.references:
                        assert set(names) == set(e.names)
                        break
                else:
                    raise ValueError(f"Could not find element: {names[0]}")

        assert set(elementlist[cell]) == set(elements)
    assert set(elementlist.keys()) == set(cells)


@pytest.mark.parametrize("script, output", doc_data)
def test_snippets(script, output):
    root = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")
    if not os.path.isfile(os.path.join(root, "VERSION")):
        # Skip test if running in tarball source
        pytest.skip()

    old_stdout = sys.stdout
    redirected_output = sys.stdout = StringIO()
    exec(script)
    sys.stdout = old_stdout

    actual_output = redirected_output.getvalue().strip()
    actual_output = re.sub(r"at 0x[^\>]+\>", "at 0x{ADDRESS}>", actual_output)
    assert actual_output == output


def test_version_numbers():
    root = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")
    if not os.path.isfile(os.path.join(root, "VERSION")):
        # Skip test if running in tarball source
        pytest.skip()

    with open(os.path.join(root, "VERSION")) as f:
        version = f.read()
    assert version == version.strip()

    # codemeta.json
    with open(os.path.join(root, "codemeta.json")) as f:
        data = json.load(f)
    assert data["version"] == version
    date = data["dateModified"]

    # setup.py
    with open(os.path.join(root, "setup.py")) as f:
        for line in f:
            if 'version="' in line:
                assert line.split('version="')[1].split('"')[0] == version

    # symfem/version.py
    with open(os.path.join(root, "symfem/version.py")) as f:
        assert f.read().split('version = "')[1].split('"')[0] == version

    # .github/workflows/test-packages.yml
    url = "https://pypi.io/packages/source/s/symfem/symfem-"
    with open(os.path.join(root, ".github/workflows/test-packages.yml")) as f:
        for line in f:
            if "ref:" in line:
                assert line.split("ref:")[1].strip() == "v" + version
            elif url in line:
                assert line.split(url)[1].split(".tar.gz")[0] == version
            elif "cd symfem-" in line:
                assert line.split("cd symfem-")[1].strip() == version

    # CITATION.cff
    with open(os.path.join(root, "CITATION.cff")) as f:
        for line in f:
            if line.startswith("version: "):
                assert line.split("version: ")[1].strip() == version
            elif line.startswith("date-released: "):
                assert line.split("date-released: ")[1].strip() == date



"""Basis function classes."""

from __future__ import annotations

import typing
from abc import ABC, abstractmethod

import sympy

from .geometry import PointType
from .references import Reference
from .symbols import AxisVariables, AxisVariablesNotSingle, t, x

SingleSympyFormat = typing.Union[
    sympy.core.expr.Expr,
    typing.Tuple[sympy.core.expr.Expr, ...],
    sympy.matrices.dense.MutableDenseMatrix
]
SympyFormat = typing.Union[
    SingleSympyFormat,
    typing.Dict[typing.Tuple[typing.Tuple[sympy.core.expr.Expr, ...], ...], SingleSympyFormat]
]
_ValuesToSubstitute = typing.Union[
    typing.Tuple[typing.Any, ...],
    typing.List[typing.Any],
    typing.Any]


def _to_sympy_format(item: typing.Any) -> SympyFormat:
    """Convert to Sympy format used by these functions.

    Args:
        item: The input item

    Returns:
        The item in Sympy format expected by functions
    """
    if isinstance(item, AnyFunction):
        return item.as_sympy()

    if isinstance(item, int):
        return sympy.Integer(item)
    if isinstance(item, sympy.core.expr.Expr):
        return item

    if isinstance(item, (list, tuple)):
        if isinstance(item[0], (list, tuple)):
            return sympy.Matrix(item)
        out = []
        for i in item:
            ii = _to_sympy_format(i)
            assert isinstance(ii, sympy.core.expr.Expr)
            out.append(ii)
        return tuple(out)

    if isinstance(item, sympy.matrices.dense.MutableDenseMatrix):
        return item

    raise NotImplementedError()


def _check_equal(first: SympyFormat, second: SympyFormat) -> bool:
    """Check if two items are equal.

    Args:
        first: The first item
        second: The second item

    Returns:
        Are the two items equal?
    """
    if isinstance(first, sympy.core.expr.Expr) and isinstance(second, sympy.core.expr.Expr):
        return (first - second).expand().simplify() == 0

    if isinstance(first, tuple) and isinstance(second, tuple):
        if len(first) != len(second):
            return False
        for i, j in zip(first, second):
            if not _check_equal(i, j):
                return False
        return True

    if isinstance(
        first, sympy.matrices.dense.MutableDenseMatrix
    ) and isinstance(
        second, sympy.matrices.dense.MutableDenseMatrix
    ):
        if first.rows != second.rows:
            return False
        if first.cols != second.cols:
            return False
        for ii in range(first.rows):
            for jj in range(first.cols):
                if not _check_equal(first[ii, jj], second[ii, jj]):
                    return False
        return True

    return False


class AnyFunction(ABC):
    """A function."""

    def __init__(self, scalar: bool = False, vector: bool = False, matrix: bool = False):
        """Create a function.

        Args:
            scalar: Is the function scalar-valued?
            vector: Is the function vector-valued?
            matrix: Is the function matrix-valued?
        """
        assert len([i for i in [scalar, vector, matrix] if i]) == 1
        self.is_scalar = scalar
        self.is_vector = vector
        self.is_matrix = matrix

    @abstractmethod
    def __add__(self, other: typing.Any):
        """Add."""
        pass

    @abstractmethod
    def __radd__(self, other: typing.Any):
        """Add."""
        pass

    @abstractmethod
    def __sub__(self, other: typing.Any):
        """Subtract."""
        pass

    @abstractmethod
    def __rsub__(self, other: typing.Any):
        """Subtract."""
        pass

    @abstractmethod
    def __neg__(self):
        """Negate."""
        pass

    @abstractmethod
    def __truediv__(self, other: typing.Any):
        """Divide."""
        pass

    @abstractmethod
    def __rtruediv__(self, other: typing.Any):
        """Divide."""
        pass

    @abstractmethod
    def __mul__(self, other: typing.Any):
        """Multiply."""
        pass

    @abstractmethod
    def __rmul__(self, other: typing.Any):
        """Multiply."""
        pass

    @abstractmethod
    def __matmul__(self, other: typing.Any):
        """Multiply."""
        pass

    @abstractmethod
    def __rmatmul__(self, other: typing.Any):
        """Multiply."""
        pass

    @abstractmethod
    def __pow__(self, other: typing.Any):
        """Raise to a power."""
        pass

    @abstractmethod
    def as_sympy(self) -> SympyFormat:
        """Convert to a Sympy expression.

        Returns:
            A Sympy expression
        """
        pass

    @abstractmethod
    def as_tex(self) -> str:
        """Convert to a TeX expression.

        Returns:
            A TeX string
        """
        pass

    @abstractmethod
    def subs(self, vars: AxisVariables, values: typing.Union[AnyFunction, _ValuesToSubstitute]):
        """Substitute values into the function.

        Args:
            vars: The variables to substitute out
            values: The value to substitute in

        Returns:
            The substituted function
        """
        pass

    @abstractmethod
    def diff(self, variable: sympy.core.symbol.Symbol):
        """Differentiate the function.

        Args:
            variable: The variable to differentiate with respect to

        Returns:
            The differentiated function
        """
        pass

    @abstractmethod
    def directional_derivative(self, direction: PointType):
        """Compute a directional derivative.

        Args:
            direction: The diection

        Returns:
            The directional differentiate
        """
        pass

    @abstractmethod
    def jacobian_component(self, component: typing.Tuple[int, int]):
        """Compute a component of the jacobian.

        Args:
            component: The component

        Returns:
            The component of the jacobian
        """
        pass

    @abstractmethod
    def jacobian(self, dim: int):
        """Compute the jacobian.

        Args:
            dim: The topological dimension of the cell

        Returns:
            The jacobian
        """
        pass

    @abstractmethod
    def dot(self, other_in: FunctionInput):
        """Compute the dot product with another function.

        Args:
            other_in: The function to multiply with

        Returns:
            The product
        """
        pass

    @abstractmethod
    def cross(self, other_in: FunctionInput):
        """Compute the cross product with another function.

        Args:
            other_in: The function to multiply with

        Returns:
            The cross product
        """
        pass

    @abstractmethod
    def div(self):
        """Compute the div of the function.

        Returns:
            The divergence
        """
        pass

    @abstractmethod
    def grad(self, dim: int):
        """Compute the grad of the function.

        Returns:
            The gradient
        """
        pass

    @abstractmethod
    def curl(self):
        """Compute the curl of the function.

        Returns:
            The curl
        """
        pass

    @abstractmethod
    def norm(self):
        """Compute the norm of the function.

        Returns:
            The norm
        """
        pass

    @abstractmethod
    def integral(self, domain: Reference, vars: AxisVariablesNotSingle = t):
        """Compute the integral of the function.

        Args:
            domain: The domain of the integral
            vars: The variables to integrate with respect to

        Returns:
            The integral
        """
        pass

    @abstractmethod
    def with_floats(self) -> AnyFunction:
        """Return a version the function with floats as coefficients.

        Returns:
            A version the function with floats as coefficients
        """
        pass

    def integrate(self, *limits: typing.Tuple[
        sympy.core.symbol.Symbol,
        typing.Union[int, sympy.core.expr.Expr],
        typing.Union[int, sympy.core.expr.Expr]
    ]):
        """Integrate the function.

        Args:
            limits: The variables and limits

        Returns:
            The integral
        """
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'integrate'")

    def det(self):
        """Compute the determinant.

        Returns:
            The deteminant
        """
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'det'")

    def transpose(self):
        """Compute the transpose.

        Returns:
            The transpose
        """
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'transpose'")

    @property
    def shape(self) -> typing.Tuple[int, ...]:
        """Get the value shape of the function.

        Returns:
            The value shape
        """
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'shape'")

    def plot(
        self, reference: Reference, filename: typing.Union[str, typing.List[str]],
        dof_point: typing.Optional[PointType] = None,
        dof_direction: typing.Optional[PointType] = None,
        dof_entity: typing.Optional[typing.Tuple[int, int]] = None,
        dof_n: typing.Optional[int] = None, value_scale: sympy.core.expr.Expr = sympy.Integer(1),
        plot_options: typing.Dict[str, typing.Any] = {}, **kwargs: typing.Any
    ):
        """Plot the function.

        Args:
            reference: The reference cell
            filename: The file name
            dof_point: The DOF point
            dof_direction: The direction of the DOF
            dof_entity: The entity the DOF is associated with
            dof_n: The number of the DOF
            value_scale: The scale factor for the function values
            plot_options: Options for the plot
            kwargs: Keyword arguments
        """
        from .plotting import Picture, colors

        extra: typing.Tuple[int, ...] = tuple()
        if self.is_scalar:
            extra = (0, )

        img = Picture(**kwargs)

        if dof_entity is not None and dof_entity[0] > 1:
            sub_e = reference.sub_entity(*dof_entity)
            img.add_fill([i + extra for i in sub_e.clockwise_vertices], colors.BLUE, 0.5)

        for ze in reference.z_ordered_entities_extra_dim():
            for dim, entity in ze:
                if dim == 1:
                    c = colors.GRAY
                    if dof_entity == (1, entity):
                        c = colors.BLUE
                    img.add_line(
                        reference.vertices[reference.edges[entity][0]] + extra,
                        reference.vertices[reference.edges[entity][1]] + extra, c)

                if dim == reference.tdim:
                    self.plot_values(reference, img, value_scale)

            if (dim, entity) in ze:
                if dof_direction is not None:
                    assert dof_point is not None and dof_n is not None
                    img.add_dof_arrow(dof_point + extra, dof_direction + extra, dof_n,
                                      colors.PURPLE, bold=False)
                elif dof_point is not None:
                    assert dof_n is not None
                    img.add_dof_marker(dof_point + extra, dof_n, colors.PURPLE, bold=False)

        img.save(filename, plot_options=plot_options)

    def plot_values(
        self, reference: Reference, img: typing.Any,
        value_scale: sympy.core.expr.Expr = sympy.Integer(1), n: int = 6
    ):
        """Plot the function's values.

        Args:
            reference: The reference cell
            img: The image to plot on
            value_scale: The scale factor for the function values
            n: The number of points per side for plotting
        """
        raise ValueError(f"Cannot plot function of type '{self.__class__.__name__}'")

    def __len__(self):
        """Compute the determinant."""
        raise TypeError(f"object of type '{self.__class__.__name__}' has no len()")

    def __getitem__(self, key) -> AnyFunction:
        """Get a component or slice of the function."""
        raise ValueError(f"'{self.__class__.__name__}' object is not subscriptable")

    def _sympy_(self) -> SympyFormat:
        """Convert to Sympy format."""
        return self.as_sympy()

    def __float__(self) -> float:
        """Convert to a float."""
        if self.is_scalar:
            a_s = self.as_sympy()
            assert isinstance(a_s, sympy.core.expr.Expr)
            return float(a_s)
        raise TypeError("Cannot convert function to a float.")

    def __lt__(self, other: typing.Any) -> bool:
        """Check inequality."""
        return self.as_sympy() < other

    def __le__(self, other: typing.Any) -> bool:
        """Check inequality."""
        return self.as_sympy() <= other

    def __gt__(self, other: typing.Any) -> bool:
        """Check inequality."""
        return self.as_sympy() > other

    def __ge__(self, other: typing.Any) -> bool:
        """Check inequality."""
        return self.as_sympy() >= other

    def __repr__(self) -> str:
        """Representation."""
        return self.as_sympy().__repr__()

    def __eq__(self, other: typing.Any) -> bool:
        """Check if two functions are equal."""
        return _check_equal(_to_sympy_format(self), _to_sympy_format(other))

    def __ne__(self, other: typing.Any) -> bool:
        """Check if two functions are not equal."""
        return not self.__eq__(other)


ValuesToSubstitute = typing.Union[AnyFunction, _ValuesToSubstitute]


class ScalarFunction(AnyFunction):
    """A scalar-valued function."""

    _f: sympy.core.expr.Expr

    def __init__(self, f: typing.Union[int, sympy.core.expr.Expr]):
        """Create a scalar-valued function.

        Args:
            f: The sympy representation of the function.
        """
        super().__init__(scalar=True)
        if isinstance(f, int):
            self._f = sympy.Integer(f)
        else:
            self._f = f
        assert isinstance(self._f, sympy.core.expr.Expr)
        self._plot_beziers: typing.Dict[typing.Tuple[Reference, int], typing.List[
            typing.Tuple[PointType, PointType, PointType, PointType]]] = {}

    def __add__(self, other: typing.Any) -> ScalarFunction:
        """Add."""
        if isinstance(other, ScalarFunction):
            return ScalarFunction(self._f + other._f)
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return ScalarFunction(self._f + other)
        return NotImplemented

    def __radd__(self, other: typing.Any) -> ScalarFunction:
        """Add."""
        if isinstance(other, ScalarFunction):
            return ScalarFunction(other._f + self._f)
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return ScalarFunction(other + self._f)
        return NotImplemented

    def __sub__(self, other: typing.Any) -> ScalarFunction:
        """Subtract."""
        if isinstance(other, ScalarFunction):
            return ScalarFunction(self._f - other._f)
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return ScalarFunction(self._f - other)
        return NotImplemented

    def __rsub__(self, other: typing.Any) -> ScalarFunction:
        """Subtract."""
        if isinstance(other, ScalarFunction):
            return ScalarFunction(other._f - self._f)
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return ScalarFunction(other - self._f)
        return NotImplemented

    def __truediv__(self, other: typing.Any) -> ScalarFunction:
        """Divide."""
        if isinstance(other, ScalarFunction):
            return ScalarFunction(self._f / other._f)
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return ScalarFunction(self._f / other)
        return NotImplemented

    def __rtruediv__(self, other: typing.Any) -> ScalarFunction:
        """Divide."""
        if isinstance(other, ScalarFunction):
            return ScalarFunction(other._f / self._f)
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return ScalarFunction(other / self._f)
        return NotImplemented

    def __mul__(self, other: typing.Any) -> ScalarFunction:
        """Multiply."""
        if isinstance(other, ScalarFunction):
            return ScalarFunction(self._f * other._f)
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return ScalarFunction(self._f * other)
        return NotImplemented

    def __rmul__(self, other: typing.Any) -> ScalarFunction:
        """Multiply."""
        if isinstance(other, ScalarFunction):
            return ScalarFunction(other._f * self._f)
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return ScalarFunction(other * self._f)
        return NotImplemented

    def __matmul__(self, other: typing.Any):
        """Multiply."""
        return NotImplemented

    def __rmatmul__(self, other: typing.Any):
        """Multiply."""
        return NotImplemented

    def __pow__(self, other: typing.Any) -> ScalarFunction:
        """Raise to a power."""
        return ScalarFunction(self._f ** other)

    def __neg__(self) -> ScalarFunction:
        """Negate."""
        return ScalarFunction(-self._f)

    def as_sympy(self) -> SympyFormat:
        """Convert to a sympy expression.

        Returns:
            A Sympy expression
        """
        return self._f

    def as_tex(self) -> str:
        """Convert to a TeX expression.

        Returns:
            A TeX string
        """
        out = sympy.latex(sympy.simplify(sympy.expand(self._f)))
        out = out.replace("\\left[", "\\left(")
        out = out.replace("\\right]", "\\right)")
        return out

    def subs(self, vars: AxisVariables, values: ValuesToSubstitute) -> ScalarFunction:
        """Substitute values into the function.

        Args:
            vars: The variables to substitute out
            values: The value to substitute in

        Returns:
            The substituted function
        """
        subbed = self._f
        if isinstance(values, AnyFunction):
            values = values.as_sympy()
        if isinstance(vars, (list, tuple)):
            assert isinstance(values, (list, tuple))
            dummy = [sympy.Symbol(f"DUMMY_{i}") for i, _ in enumerate(values)]
            for i, j in zip(vars, dummy):
                subbed = subbed.subs(i, j)
            for i, j in zip(dummy, values):
                subbed = subbed.subs(i, j)
        else:
            assert not isinstance(values, (list, tuple))
            subbed = subbed.subs(vars, values)
        return ScalarFunction(subbed)

    def diff(self, variable: sympy.core.symbol.Symbol) -> ScalarFunction:
        """Differentiate the function.

        Args:
            variable: The variable to differentiate with respect to

        Returns:
            The differentiated function
        """
        return ScalarFunction(self._f.diff(variable))

    def directional_derivative(self, direction: PointType) -> ScalarFunction:
        """Compute a directional derivative.

        Args:
            direction: The diection

        Returns:
            The directional derivatve
        """
        out = ScalarFunction(0)
        for i, j in zip(x, direction):
            out += j * self.diff(i)
        return out

    def jacobian_component(self, component: typing.Tuple[int, int]) -> ScalarFunction:
        """Compute a component of the jacobian.

        Args:
            component: The component

        Returns:
            The component of the jacobian
        """
        return self.diff(x[component[0]]).diff(x[component[1]])

    def jacobian(self, dim: int) -> MatrixFunction:
        """Compute the jacobian.

        Args:
            dim: The topological dimension of the cell

        Returns:
            The jacobian
        """
        return MatrixFunction(tuple(
            tuple(self.jacobian_component((i, j)) for j in range(dim))
            for i in range(dim)))

    def dot(self, other_in: FunctionInput) -> ScalarFunction:
        """Compute the dot product with another function.

        Args:
            other_in: The function to multiply with

        Returns:
            The product
        """
        other = parse_function_input(other_in)
        if isinstance(other, ScalarFunction):
            return self * other

        if isinstance(other, AnyFunction) and other.is_scalar:
            return other.dot(self)

        raise NotImplementedError()

    def cross(self, other_in: FunctionInput):
        """Compute the cross product with another function.

        Args:
            other_in: The function to multiply with

        Returns:
            The cross product
        """
        raise ValueError("Cannot cross a scalar-valued function.")

    def div(self):
        """Compute the div of the function.

        Returns:
            The divergence
        """
        raise ValueError("Cannot compute the div of a scalar-valued function.")

    def grad(self, dim: int) -> VectorFunction:
        """Compute the grad of the function.

        Returns:
            The gradient
        """
        return VectorFunction([self.diff(x[i]) for i in range(dim)])

    def curl(self):
        """Compute the curl of the function.

        Returns:
            The curl
        """
        raise ValueError("Cannot compute the curl of a scalar-valued function.")

    def norm(self) -> ScalarFunction:
        """Compute the norm of the function.

        Returns:
            The norm
        """
        return ScalarFunction(abs(self._f))

    def integral(self, domain: Reference, vars: AxisVariablesNotSingle = t) -> ScalarFunction:
        """Compute the integral of the function.

        Args:
            domain: The domain of the integral
            vars: The variables to integrate with respect to

        Returns:
            The integral
        """
        limits = domain.integration_limits(vars)

        point = VectorFunction(domain.origin)
        for ti, a in zip(t, domain.axes):
            point += ti * VectorFunction(a)
        out = self._f.subs(x, point)

        if len(limits[0]) == 2:
            for i in limits:
                assert len(i) == 2
                out = out.subs(*i)
            return out

        out *= domain.jacobian()
        return ScalarFunction(out.integrate(*limits))

    def integrate(self, *limits: typing.Tuple[
        sympy.core.symbol.Symbol,
        typing.Union[int, sympy.core.expr.Expr],
        typing.Union[int, sympy.core.expr.Expr]
    ]):
        """Integrate the function.

        Args:
            limits: The variables and limits

        Returns:
            The integral
        """
        return ScalarFunction(self._f.integrate(*limits))

    def plot_values(
        self, reference: Reference, img: typing.Any,
        value_scale: sympy.core.expr.Expr = sympy.Integer(1), n: int = 6
    ):
        """Plot the function's values.

        Args:
            reference: The reference cell
            img: The image to plot on
            value_scale: The scale factor for the function values
            n: The number of points per side for plotting
        """
        from .plotting import Picture, colors
        assert isinstance(img, Picture)

        if (reference, n) not in self._plot_beziers:
            self._plot_beziers[(reference, n)] = []
            pts, pairs = reference.make_lattice_with_lines_float(n)

            value_scale *= sympy.Rational(5, 8)
            value_scale = sympy.Float(float(value_scale))

            deriv = self.grad(reference.tdim)
            evals = []
            for p in pts:
                value = self.subs(x, p).as_sympy()
                assert isinstance(value, sympy.core.expr.Expr)
                value = sympy.Float(float(value))
                value *= value_scale
                evals.append(value)

            for i, j in pairs:
                pi = VectorFunction(pts[i])
                pj = VectorFunction(pts[j])
                d_pi = (2 * pi + pj) / 3
                d_pj = (2 * pj + pi) / 3
                di = deriv.subs(x, pi).dot(d_pi - pts[i]).as_sympy()
                dj = deriv.subs(x, pj).dot(d_pj - pts[j]).as_sympy()
                assert isinstance(di, sympy.core.expr.Expr)
                assert isinstance(dj, sympy.core.expr.Expr)
                self._plot_beziers[(reference, n)].append((
                    tuple(pi) + (evals[i], ), tuple(d_pi) + (evals[i] + di * value_scale, ),
                    tuple(d_pj) + (evals[j] + dj * value_scale, ), tuple(pj) + (evals[j], )))

        for s, m1, m2, e in self._plot_beziers[(reference, n)]:
            img.add_bezier(s, m1, m2, e, colors.ORANGE)

    def with_floats(self) -> AnyFunction:
        """Return a version the function with floats as coefficients.

        Returns:
            A version the function with floats as coefficients
        """
        out = sympy.Float(0.0)
        for term, co in self._f.as_coefficients_dict().items():
            out += float(co) * term
        return ScalarFunction(out)


class VectorFunction(AnyFunction):
    """A vector-valued function."""

    _vec: tuple[ScalarFunction, ...]

    def __init__(self, vec: typing.Union[
        typing.Tuple[typing.Union[AnyFunction, int, sympy.core.expr.Expr], ...],
        typing.List[typing.Union[AnyFunction, int, sympy.core.expr.Expr]]
    ]):
        """Create a vector-valued function.

        Args:
            vec: The sympy representation of the function.
        """
        from .basis_functions import BasisFunction
        super().__init__(vector=True)
        vec_l = []
        for i in vec:
            if isinstance(i, AnyFunction):
                if isinstance(i, BasisFunction):
                    i = i.get_function()
                assert isinstance(i, ScalarFunction)
                vec_l.append(i)
            else:
                vec_l.append(ScalarFunction(i))
        self._vec = tuple(vec_l)
        for i in self._vec:
            assert i.is_scalar

        self._plot_arrows: typing.Dict[typing.Tuple[Reference, int], typing.List[
            typing.Tuple[typing.Tuple[sympy.core.expr.Expr, ...], VectorFunction, float]]] = {}

    def __len__(self):
        """Get the length of the vector."""
        return len(self._vec)

    @property
    def shape(self) -> typing.Tuple[int, ...]:
        """Get the value shape of the function.

        Returns:
            The value shape
        """
        return (len(self), )

    def __getitem__(self, key) -> typing.Union[ScalarFunction, VectorFunction]:
        """Get a component or slice of the function."""
        fs = self._vec[key]
        if isinstance(fs, ScalarFunction):
            return fs
        return VectorFunction(fs)

    def __add__(self, other: typing.Any) -> VectorFunction:
        """Add."""
        if isinstance(other, VectorFunction):
            assert len(self._vec) == len(other._vec)
            return VectorFunction(tuple(i._f + j._f for i, j in zip(self._vec, other._vec)))
        if isinstance(other, (list, tuple)):
            return VectorFunction(tuple(i._f + j for i, j in zip(self._vec, other)))
        return NotImplemented

    def __radd__(self, other: typing.Any) -> VectorFunction:
        """Add."""
        if isinstance(other, VectorFunction):
            assert len(self._vec) == len(other._vec)
            return VectorFunction(tuple(i._f + j._f for i, j in zip(other._vec, self._vec)))
        if isinstance(other, (list, tuple)):
            return VectorFunction(tuple(i + j._f for i, j in zip(other, self._vec)))
        return NotImplemented

    def __sub__(self, other: typing.Any) -> VectorFunction:
        """Subtract."""
        if isinstance(other, VectorFunction):
            assert len(self._vec) == len(other._vec)
            return VectorFunction(tuple(i._f - j._f for i, j in zip(self._vec, other._vec)))
        if isinstance(other, (list, tuple)):
            return VectorFunction(tuple(i._f - j for i, j in zip(self._vec, other)))
        return NotImplemented

    def __rsub__(self, other: typing.Any) -> VectorFunction:
        """Subtract."""
        if isinstance(other, VectorFunction):
            assert len(self._vec) == len(other._vec)
            return VectorFunction(tuple(i._f - j._f for i, j in zip(other._vec, self._vec)))
        if isinstance(other, (list, tuple)):
            return VectorFunction(tuple(i - j._f for i, j in zip(other, self._vec)))
        return NotImplemented

    def __neg__(self) -> VectorFunction:
        """Negate."""
        return VectorFunction(tuple(-i._f for i in self._vec))

    def __truediv__(self, other: typing.Any) -> VectorFunction:
        """Divide."""
        if isinstance(other, ScalarFunction):
            return VectorFunction(tuple(i._f / other._f for i in self._vec))
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return VectorFunction(tuple(i._f / other for i in self._vec))
        return NotImplemented

    def __rtruediv__(self, other: typing.Any) -> VectorFunction:
        """Divide."""
        return NotImplemented

    def __mul__(self, other: typing.Any) -> VectorFunction:
        """Multiply."""
        if isinstance(other, ScalarFunction):
            return VectorFunction(tuple(i._f * other._f for i in self._vec))
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return VectorFunction(tuple(i._f * other for i in self._vec))
        return NotImplemented

    def __rmul__(self, other: typing.Any) -> VectorFunction:
        """Multiply."""
        if isinstance(other, ScalarFunction):
            return VectorFunction(tuple(other._f * i._f for i in self._vec))
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return VectorFunction(tuple(other * i._f for i in self._vec))
        return NotImplemented

    def __matmul__(self, other: typing.Any) -> VectorFunction:
        """Multiply."""
        if isinstance(other, MatrixFunction):
            assert other.shape[1] == len(self)
            return VectorFunction([other.col(i).dot(self) for i in range(other.shape[0])])
        return NotImplemented

    def __rmatmul__(self, other: typing.Any) -> VectorFunction:
        """Multiply."""
        if isinstance(other, MatrixFunction):
            assert other.shape[1] == len(self)
            return VectorFunction([other.row(i).dot(self) for i in range(other.shape[0])])
        return NotImplemented

    def __pow__(self, other: typing.Any) -> VectorFunction:
        """Raise to a power."""
        if isinstance(other, ScalarFunction):
            return VectorFunction(tuple(i._f ** other._f for i in self._vec))
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return VectorFunction(tuple(i._f ** other for i in self._vec))
        return NotImplemented

    def as_sympy(self) -> SympyFormat:
        """Convert to a sympy expression.

        Returns:
            A Sympy expression
        """
        return tuple(i._f for i in self._vec)

    def as_tex(self) -> str:
        """Convert to a TeX expression.

        Returns:
            A TeX string
        """
        return "\\left(\\begin{array}{c}" + "\\\\".join([
            "\\displaystyle " + i.as_tex() for i in self._vec
        ]) + "\\end{array}\\right)"

    def subs(self, vars: AxisVariables, values: ValuesToSubstitute) -> VectorFunction:
        """Substitute values into the function.

        Args:
            vars: The variables to substitute out
            values: The value to substitute in

        Returns:
            The substituted function
        """
        subbed = tuple(i.subs(vars, values) for i in self._vec)
        return VectorFunction(subbed)

    def diff(self, variable: sympy.core.symbol.Symbol) -> VectorFunction:
        """Differentiate the function.

        Args:
            variable: The variable to differentiate with respect to

        Returns:
            The differentiated function
        """
        return VectorFunction([i.diff(variable) for i in self._vec])

    def directional_derivative(self, direction: PointType):
        """Compute a directional derivative.

        Args:
            direction: The diection

        Returns:
            The directional derivatve
        """
        raise NotImplementedError()

    def jacobian_component(self, component: typing.Tuple[int, int]):
        """Compute a component of the jacobian.

        Args:
            component: The component

        Returns:
            The component of the jacobian
        """
        raise NotImplementedError()

    def jacobian(self, dim: int) -> MatrixFunction:
        """Compute the jacobian.

        Args:
            dim: The topological dimension of the cell

        Returns:
            The jacobian
        """
        raise NotImplementedError()

    def dot(self, other_in: FunctionInput) -> ScalarFunction:
        """Compute the dot product with another function.

        Args:
            other_in: The function to multiply with

        Returns:
            The product
        """
        other = parse_function_input(other_in)
        if isinstance(other, VectorFunction):
            assert len(self._vec) == len(other._vec)
            out = 0
            for i, j in zip(self._vec, other._vec):
                out += i._f * j._f
            return ScalarFunction(out)

        if isinstance(other, AnyFunction) and other.is_vector:
            return other.dot(self)

        raise NotImplementedError()

    def cross(self, other_in: FunctionInput) -> typing.Union[VectorFunction, ScalarFunction]:
        """Compute the cross product with another function.

        Args:
            other_in: The function to multiply with

        Returns:
            The cross product
        """
        other = parse_function_input(other_in)
        assert other.is_vector and len(self) == len(other)
        if len(self) == 2:
            return self[0] * other[1] - self[1] * other[0]
        else:
            assert len(self) == 3
            return VectorFunction([self[1] * other[2] - self[2] * other[1],
                                   self[2] * other[0] - self[0] * other[2],
                                   self[0] * other[1] - self[1] * other[0]])

    def div(self) -> ScalarFunction:
        """Compute the div of the function.

        Returns:
            The divergence
        """
        out = ScalarFunction(0)
        for i, j in zip(self._vec, x):
            out += i.diff(j)
        return out

    def grad(self):
        """Compute the grad of the function.

        Returns:
            The gradient
        """
        raise ValueError("Cannot compute the grad of a vector-valued function.")

    def curl(self) -> VectorFunction:
        """Compute the curl of the function.

        Returns:
            The curl
        """
        assert len(self._vec) == 3
        return VectorFunction([
            self._vec[2].diff(x[1]) - self._vec[1].diff(x[2]),
            self._vec[0].diff(x[2]) - self._vec[2].diff(x[0]),
            self._vec[1].diff(x[0]) - self._vec[0].diff(x[1])
        ])

    def norm(self) -> ScalarFunction:
        """Compute the norm of the function.

        Returns:
            The norm
        """
        a = sympy.Integer(0)
        for i in self._vec:
            a += i._f ** 2
        return ScalarFunction(sympy.sqrt(a))

    def integral(self, domain: Reference, vars: AxisVariablesNotSingle = t):
        """Compute the integral of the function.

        Args:
            domain: The domain of the integral
            vars: The variables to integrate with respect to

        Returns:
            The integral
        """
        raise NotImplementedError()

    def __iter__(self):
        """Get iterable."""
        self.iter_n = 0
        return self

    def __next__(self):
        """Get next item."""
        if self.iter_n < len(self._vec):
            self.iter_n += 1
            return self._vec[self.iter_n - 1]
        else:
            raise StopIteration

    def plot_values(
        self, reference: Reference, img: typing.Any,
        value_scale: sympy.core.expr.Expr = sympy.Integer(1), n: int = 6
    ):
        """Plot the function's values.

        Args:
            reference: The reference cell
            img: The image to plot on
            value_scale: The scale factor for the function values
            n: The number of points per side for plotting
        """
        from .plotting import Picture, colors
        assert isinstance(img, Picture)

        if (reference, n) not in self._plot_arrows:
            self._plot_arrows[(reference, n)] = []
            pts = reference.make_lattice_float(n)
            value_scale /= 4

            for p in pts:
                value_s = self.subs(x, p).as_sympy()
                assert isinstance(value_s, tuple)
                value = VectorFunction(tuple(sympy.Float(float(v)) for v in value_s))
                value *= value_scale
                size = float(value.norm() * 40)
                self._plot_arrows[(reference, n)].append((p, VectorFunction(p) + value, size))
        for p, q, size in self._plot_arrows[(reference, n)]:
            img.add_arrow(p, q, colors.ORANGE, size)

    def with_floats(self) -> AnyFunction:
        """Return a version the function with floats as coefficients.

        Returns:
            A version the function with floats as coefficients
        """
        return VectorFunction(tuple(f.with_floats() for f in self._vec))


class MatrixFunction(AnyFunction):
    """A matrix-valued function."""

    _mat: typing.Tuple[typing.Tuple[ScalarFunction, ...], ...]

    def __init__(self, mat: typing.Union[
        typing.Tuple[typing.Tuple[typing.Union[AnyFunction, int, sympy.core.expr.Expr],
                                  ...], ...],
        typing.Tuple[typing.List[typing.Union[AnyFunction, int, sympy.core.expr.Expr]], ...],
        typing.List[typing.Tuple[typing.Union[AnyFunction, int, sympy.core.expr.Expr], ...]],
        typing.List[typing.List[typing.Union[AnyFunction, int, sympy.core.expr.Expr]]],
        sympy.matrices.dense.MutableDenseMatrix
    ]):
        """Create a matrix-valued function.

        Args:
            mat: The sympy representation of the function.
        """
        from .basis_functions import BasisFunction
        super().__init__(matrix=True)
        if isinstance(mat, sympy.matrices.dense.MutableDenseMatrix):
            mat = tuple(tuple(mat[i, j] for j in range(mat.cols)) for i in range(mat.rows))
        assert isinstance(mat, (list, tuple))
        mat_l = []
        for i in mat:
            row = []
            for j in i:
                if isinstance(j, AnyFunction):
                    if isinstance(j, BasisFunction):
                        j = j.get_function()
                    assert isinstance(j, ScalarFunction)
                    row.append(j)
                else:
                    row.append(ScalarFunction(j))
            mat_l.append(tuple(row))

        self._mat = tuple(mat_l)
        self._shape = (len(self._mat), 0 if len(self._mat) == 0 else len(self._mat[0]))
        for i in self._mat:
            assert len(i) == self._shape[1]
            for j in i:
                assert j.is_scalar

    @property
    def shape(self) -> typing.Tuple[int, ...]:
        """Get the value shape of the function.

        Returns:
            The value shape
        """
        return self._shape

    def __getitem__(self, key) -> typing.Union[ScalarFunction, VectorFunction]:
        """Get a component or slice of the function."""
        if isinstance(key, tuple):
            assert len(key) == 2
            return self._mat[key[0]][key[1]]
        return self.row(key)

    def row(self, n: int) -> VectorFunction:
        """Get a row of the matrix.

        Args:
            n: The row number

        Returns:
            The row of the matrix
        """
        return VectorFunction([self._mat[n][i] for i in range(self.shape[1])])

    def col(self, n: int) -> VectorFunction:
        """Get a colunm of the matrix.

        Args:
            n: The column number

        Returns:
            The column of the matrix
        """
        return VectorFunction([self._mat[i][n] for i in range(self.shape[0])])

    def __add__(self, other: typing.Any) -> MatrixFunction:
        """Add."""
        if isinstance(other, MatrixFunction):
            assert self.shape == other.shape
            return MatrixFunction(tuple(tuple(
                ii._f + jj._f for ii, jj in zip(i, j)) for i, j in zip(self._mat, other._mat)))
        if isinstance(other, (list, tuple)):
            data = []
            for i, j in zip(self._mat, other):
                assert isinstance(j, (list, tuple)) and len(j) == self.shape[1]
                data.append(tuple(ii._f + jj for ii, jj in zip(i, j)))
            return MatrixFunction(data)
        if isinstance(other, sympy.matrices.dense.MutableDenseMatrix):
            assert other.rows == self.shape[0]
            assert other.cols == self.shape[1]
            return MatrixFunction([[self._mat[i][j]._f + other[i, j]
                                    for j in range(self.shape[1])] for i in range(self.shape[0])])
        return NotImplemented

    def __radd__(self, other: typing.Any) -> MatrixFunction:
        """Add."""
        if isinstance(other, MatrixFunction):
            assert self.shape == other.shape
            return MatrixFunction(tuple(tuple(
                ii._f + jj._f for ii, jj in zip(i, j)) for i, j in zip(other._mat, self._mat)))
        if isinstance(other, (list, tuple)):
            data = []
            for i, j in zip(other, self._mat):
                assert isinstance(i, (list, tuple)) and len(i) == self.shape[1]
                data.append(tuple(ii + jj._f for ii, jj in zip(i, j)))
            return MatrixFunction(data)
        if isinstance(other, sympy.matrices.dense.MutableDenseMatrix):
            assert other.rows == self.shape[0]
            assert other.cols == self.shape[1]
            return MatrixFunction([[other[i, j] + self._mat[i][j]._f
                                    for j in range(self.shape[1])] for i in range(self.shape[0])])
        return NotImplemented

    def __sub__(self, other: typing.Any) -> MatrixFunction:
        """Subtract."""
        if isinstance(other, MatrixFunction):
            assert self.shape == other.shape
            return MatrixFunction(tuple(tuple(
                ii._f - jj._f for ii, jj in zip(i, j)) for i, j in zip(self._mat, other._mat)))
        if isinstance(other, (list, tuple)):
            data = []
            for i, j in zip(self._mat, other):
                assert isinstance(j, (list, tuple)) and len(j) == self.shape[1]
                data.append(tuple(ii._f - jj for ii, jj in zip(i, j)))
            return MatrixFunction(data)
        if isinstance(other, sympy.matrices.dense.MutableDenseMatrix):
            assert other.rows == self.shape[0]
            assert other.cols == self.shape[1]
            return MatrixFunction([[self._mat[i][j]._f - other[i, j]
                                    for j in range(self.shape[1])] for i in range(self.shape[0])])
        return NotImplemented

    def __rsub__(self, other: typing.Any) -> MatrixFunction:
        """Subtract."""
        if isinstance(other, MatrixFunction):
            assert self.shape == other.shape
            return MatrixFunction(tuple(tuple(
                ii._f - jj._f for ii, jj in zip(i, j)) for i, j in zip(other._mat, self._mat)))
        if isinstance(other, (list, tuple)):
            data = []
            for i, j in zip(other, self._mat):
                assert isinstance(i, (list, tuple)) and len(i) == self.shape[1]
                data.append(tuple(ii - jj._f for ii, jj in zip(i, j)))
            return MatrixFunction(data)
        if isinstance(other, sympy.matrices.dense.MutableDenseMatrix):
            assert other.rows == self.shape[0]
            assert other.cols == self.shape[1]
            return MatrixFunction([[other[i, j] - self._mat[i][j]._f
                                    for j in range(self.shape[1])] for i in range(self.shape[0])])
        return NotImplemented

    def __neg__(self) -> MatrixFunction:
        """Negate."""
        return MatrixFunction(tuple(tuple(-j._f for j in i) for i in self._mat))

    def __truediv__(self, other: typing.Any) -> MatrixFunction:
        """Divide."""
        if isinstance(other, ScalarFunction):
            return MatrixFunction(tuple(tuple(j._f / other._f for j in i) for i in self._mat))
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return MatrixFunction(tuple(tuple(j._f / other for j in i) for i in self._mat))
        return NotImplemented

    def __rtruediv__(self, other: typing.Any) -> MatrixFunction:
        """Divide."""
        return NotImplemented

    def __mul__(self, other: typing.Any) -> MatrixFunction:
        """Multiply."""
        if isinstance(other, ScalarFunction):
            return MatrixFunction(tuple(tuple(j._f * other._f for j in i) for i in self._mat))
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return MatrixFunction(tuple(tuple(j._f * other for j in i) for i in self._mat))
        return NotImplemented

    def __rmul__(self, other: typing.Any) -> MatrixFunction:
        """Multiply."""
        if isinstance(other, ScalarFunction):
            return MatrixFunction(tuple(tuple(other._f * j._f for j in i) for i in self._mat))
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return MatrixFunction(tuple(tuple(other * j._f for j in i) for i in self._mat))
        return NotImplemented

    def __matmul__(self, other: typing.Any) -> MatrixFunction:
        """Multiply."""
        if isinstance(other, MatrixFunction):
            assert other.shape[0] == self.shape[1]
            return MatrixFunction(tuple(
                tuple(self.row(i).dot(other.col(j)) for j in range(other.shape[1]))
                for i in range(self.shape[0])))
        return NotImplemented

    def __rmatmul__(self, other: typing.Any) -> MatrixFunction:
        """Multiply."""
        if isinstance(other, MatrixFunction):
            assert self.shape[0] == other.shape[1]
            return MatrixFunction(tuple(
                tuple(other.row(i).dot(self.col(j)) for j in range(self.shape[1]))
                for i in range(other.shape[0])))
        return NotImplemented

    def __pow__(self, other: typing.Any) -> MatrixFunction:
        """Raise to a power."""
        if isinstance(other, ScalarFunction):
            return MatrixFunction(tuple(tuple(j._f ** other._f for j in i) for i in self._mat))
        if isinstance(other, (int, sympy.core.expr.Expr)):
            return MatrixFunction(tuple(tuple(j._f ** other for j in i) for i in self._mat))
        return NotImplemented

    def as_sympy(self) -> SympyFormat:
        """Convert to a sympy expression.

        Returns:
            A Sympy expression
        """
        return sympy.Matrix([[j._f for j in i] for i in self._mat])

    def as_tex(self) -> str:
        """Convert to a TeX expression.

        Returns:
            A TeX string
        """
        out = "\\left(\\begin{array}{"
        out += "c" * self.shape[1]
        out += "}"
        out += "\\\\".join([
            "&".join(["\\displaystyle" + j.as_tex() for j in i])
            for i in self._mat
        ])
        out += "\\end{array}\\right)"
        return out

    def subs(self, vars: AxisVariables, values: ValuesToSubstitute) -> MatrixFunction:
        """Substitute values into the function.

        Args:
            vars: The variables to substitute out
            values: The value to substitute in

        Returns:
            The substituted function
        """
        subbed = tuple(tuple(j.subs(vars, values) for j in i) for i in self._mat)
        return MatrixFunction(subbed)

    def diff(self, variable: sympy.core.symbol.Symbol) -> MatrixFunction:
        """Differentiate the function.

        Args:
            variable: The variable to differentiate with respect to

        Returns:
            The differentiated function
        """
        return MatrixFunction(tuple(
            tuple(self._mat[i][j].diff(variable) for j in range(self.shape[1]))
            for i in range(self.shape[0])))

    def directional_derivative(self, direction: PointType):
        """Compute a directional derivative.

        Args:
            direction: The diection

        Returns:
            The directional derivatve
        """
        raise NotImplementedError()

    def jacobian_component(self, component: typing.Tuple[int, int]):
        """Compute a component of the jacobian.

        Args:
            component: The component

        Returns:
            The component of the jacobian
        """
        raise NotImplementedError()

    def jacobian(self, dim: int):
        """Compute the jacobian.

        Args:
            dim: The topological dimension of the cell

        Returns:
            The jacobian
        """
        raise NotImplementedError()

    def dot(self, other_in: FunctionInput) -> ScalarFunction:
        """Compute the dot product with another function.

        Args:
            other_in: The function to multiply with

        Returns:
            The product
        """
        other = parse_function_input(other_in)
        if isinstance(other, MatrixFunction):
            assert self.shape == other.shape
            out = ScalarFunction(0)
            for i in range(self.shape[0]):
                for j in range(self.shape[0]):
                    out += self._mat[i][j] * other._mat[i][j]
            return out

        if isinstance(other, AnyFunction) and other.is_matrix:
            return other.dot(self)

        raise NotImplementedError()

    def cross(self, other_in: FunctionInput):
        """Compute the cross product with another function.

        Args:
            other_in: The function to multiply with

        Returns:
            The cross product
        """
        raise ValueError("Cannot cross a matrix-valued function.")

    def div(self):
        """Compute the div of the function.

        Returns:
            The divergence
        """
        raise ValueError("Cannot compute the div of a matrix-valued function.")

    def grad(self):
        """Compute the grad of the function.

        Returns:
            The gradient
        """
        raise ValueError("Cannot compute the grad of a matrix-valued function.")

    def curl(self):
        """Compute the curl of the function.

        Returns:
            The curl
        """
        raise ValueError("Cannot compute the curl of a matrix-valued function.")

    def norm(self) -> ScalarFunction:
        """Compute the norm of the function.

        Returns:
            The norm
        """
        raise NotImplementedError()

    def integral(self, domain: Reference, vars: AxisVariablesNotSingle = t):
        """Compute the integral of the function.

        Args:
            domain: The domain of the integral
            vars: The variables to integrate with respect to

        Returns:
            The integral
        """
        raise NotImplementedError()

    def det(self) -> ScalarFunction:
        """Compute the determinant.

        Returns:
            The deteminant
        """
        if self.shape[0] == self.shape[1]:
            mat = self.as_sympy()
            assert isinstance(mat, sympy.matrices.dense.MutableDenseMatrix)
            return ScalarFunction(mat.det())
        if self.shape[0] == 3 and self.shape[1] == 2:
            return self.col(0).cross(self.col(1)).norm()
        raise ValueError(f"Cannot find determinant of {self.shape[0]}x{self.shape[1]} matrix.")

    def transpose(self) -> MatrixFunction:
        """Compute the transpose.

        Returns:
            The transpose
        """
        mat = self.as_sympy()
        assert isinstance(mat, sympy.matrices.dense.MutableDenseMatrix)
        return MatrixFunction(mat.transpose())

    def with_floats(self) -> AnyFunction:
        """Return a version the function with floats as coefficients.

        Returns:
            A version the function with floats as coefficients
        """
        return MatrixFunction(tuple(tuple(f.with_floats() for f in row) for row in self._mat))


FunctionInput = typing.Union[
    AnyFunction,
    sympy.core.expr.Expr, int,
    typing.Tuple[typing.Union[sympy.core.expr.Expr, int, AnyFunction], ...],
    typing.List[typing.Union[sympy.core.expr.Expr, int, AnyFunction]],
    typing.Tuple[typing.Tuple[typing.Union[sympy.core.expr.Expr, int, AnyFunction], ...], ...],
    typing.Tuple[typing.List[typing.Union[sympy.core.expr.Expr, int, AnyFunction]], ...],
    typing.List[typing.Tuple[typing.Union[sympy.core.expr.Expr, int, AnyFunction], ...]],
    typing.List[typing.List[typing.Union[sympy.core.expr.Expr, int, AnyFunction]]],
    sympy.matrices.dense.MutableDenseMatrix]


def parse_function_input(f: FunctionInput) -> AnyFunction:
    """Parse a function.

    Args:
        f: A function

    Returns:
        The function as a Symfem function
    """
    if isinstance(f, AnyFunction):
        return f
    if isinstance(f, (sympy.core.expr.Expr, int)):
        return ScalarFunction(f)
    if isinstance(f, (list, tuple)):
        if isinstance(f[0], (list, tuple)):
            mat = []
            for i in f:
                assert isinstance(i, (tuple, list))
                mat.append(tuple(i))
            return MatrixFunction(tuple(mat))
        else:
            vec = []
            for i in f:
                assert not isinstance(i, (tuple, list))
                vec.append(i)
            return VectorFunction(tuple(vec))
    raise ValueError(f"Could not parse input function: {f}")


def parse_function_list_input(
    functions: typing.Union[typing.List[FunctionInput], typing.Tuple[FunctionInput, ...]]
) -> typing.List[AnyFunction]:
    """Parse a list of functions.

    Args:
        functions: The functions

    Returns:
        The functions as Symfem functions
    """
    return [parse_function_input(f) for f in functions]



"""Functionals used to define the dual sets."""

import typing
from abc import ABC, abstractmethod

import sympy

from . import mappings
from .functions import (AnyFunction, FunctionInput, ScalarFunction, VectorFunction,
                        parse_function_input)
from .geometry import PointType, SetOfPoints
from .references import Interval, Reference
from .symbols import t, x

ScalarValueOrFloat = typing.Union[sympy.core.expr.Expr, float]


def _to_tex(f: FunctionInput, tfrac: bool = False) -> str:
    r"""Convert an expresson to TeX.

    Args:
        f: the function
        tfrac: Should \\tfrac be used in the place of \\frac?

    Returns:
        The function as TeX
    """
    out = parse_function_input(f).as_tex()

    if tfrac:
        return out.replace("\\frac", "\\tfrac")
    else:
        return out


def _nth(n: int) -> str:
    """Add th, st or nd to a number.

    Args:
        n: The number

    Returns:
        The string (n)th, (n)st or (n)nd
    """
    if n % 10 == 1 and n != 11:
        return f"{n}st"
    if n % 10 == 2 and n != 12:
        return f"{n}nd"
    return f"{n}th"


class BaseFunctional(ABC):
    """A functional."""

    def __init__(self, reference: Reference, entity: typing.Tuple[int, int],
                 mapping: typing.Union[str, None]):
        """Create the functional.

        Args:
            reference: The reference cell
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        self.reference = reference
        self.entity = entity
        self.mapping = mapping

    def entity_dim(self) -> int:
        """Get the dimension of the entity this DOF is associated with.

        Returns:
            The dimension of the entity this DOF is associated with
        """
        return self.entity[0]

    def entity_number(self) -> int:
        """Get the number of the entity this DOF is associated with.

        Returns:
            The number of the entity this DOF is associated with
        """
        return self.entity[1]

    def perform_mapping(
        self, fs: typing.List[AnyFunction], map: PointType, inverse_map: PointType, tdim: int
    ) -> typing.List[AnyFunction]:
        """Map functions to a cell.

        Args:
            fs: functions
            map: Map from the reference cell to a physical cell
            inverse_map: Map to the reference cell from a physical cell
            tdim: The topological dimension of the cell

        Returns:
            Mapped functions
        """
        assert self.mapping is not None
        return [getattr(mappings, self.mapping)(f, map, inverse_map, tdim) for f in fs]

    def entity_tex(self) -> str:
        """Get the entity the DOF is associated with in TeX format.

        Returns:
            TeX representation of entity
        """
        if self.entity[0] == self.reference.tdim:
            return "R"
        else:
            return f"{'vefc'[self.entity[0]]}_{{{self.entity[1]}}}"

    def entity_definition(self) -> str:
        """Get the definition of the entity the DOF is associated with.

        Returns:
            The definition
        """
        if self.entity[0] == self.reference.tdim:
            return "\\(R\\) is the reference element"
        else:
            desc = f"\\({self.entity_tex()}\\) is the {_nth(self.entity[1])} "
            desc += ['vertex', 'edge', 'face', 'volume'][self.entity[0]]
            return desc

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        return None

    def eval(self, function: AnyFunction, symbolic: bool = True) -> typing.Union[
        ScalarFunction, float
    ]:
        """Apply to the functional to a function.

        Args:
            function: The function
            symbolic: Should it be applied symbolically?

        Returns:
            The value of the functional for the function
        """
        value = self.eval_symbolic(function)
        if symbolic:
            return value
        else:
            return float(value)

    @abstractmethod
    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pass

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        point = self.dof_point()
        if self.entity[0] == 0:
            return point

        midpoint = self.reference.sub_entity(*self.entity).midpoint()
        return tuple(
            m + sympy.Rational(7, 10) * (p - m)
            for m, p in zip(midpoint, point))

    def eval_symbolic(self, function: AnyFunction) -> ScalarFunction:
        """Symbolically apply the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        e = self._eval_symbolic(function)
        assert isinstance(e, ScalarFunction)
        es = e.as_sympy()
        assert isinstance(es, sympy.core.expr.Expr)
        assert es.is_constant()
        return e

    @abstractmethod
    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        pass

    @abstractmethod
    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        pass

    name = "Base functional"


class PointEvaluation(BaseFunctional):
    """A point evaluation."""

    def __init__(self, reference: Reference, point_in: FunctionInput,
                 entity: typing.Tuple[int, int], mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The point
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        return function.subs(x, self.point)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        assert isinstance(self.point, VectorFunction)
        return f"v\\mapsto v({','.join([_to_tex(i, True) for i in self.point])})", []

    name = "Point evaluation"


class WeightedPointEvaluation(BaseFunctional):
    """A point evaluation."""

    def __init__(self, reference: Reference, point_in: FunctionInput, weight: sympy.core.expr.Expr,
                 entity: typing.Tuple[int, int], mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The point
            weight: The weight
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector
        self.weight = weight

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        return function.subs(x, self.point) * self.weight

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        assert isinstance(self.point, VectorFunction)
        return (f"v\\mapsto {_to_tex(self.weight)} "
                f"v({','.join([_to_tex(i, True) for i in self.point])})"), []

    name = "Weighted point evaluation"


class DerivativePointEvaluation(BaseFunctional):
    """A point evaluation of a given derivative."""

    def __init__(self, reference: Reference, point_in: FunctionInput,
                 derivative: typing.Tuple[int, ...],
                 entity: typing.Tuple[int, int], mapping: typing.Optional[str] = None):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The point
            derivative: The order(s) of the derivative
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector
        self.derivative = derivative

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        for i, j in zip(x, self.derivative):
            for k in range(j):
                function = function.diff(i)
        return function.subs(x, self.point)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def perform_mapping(
        self, fs: typing.List[AnyFunction], map: PointType, inverse_map: PointType, tdim: int
    ) -> typing.List[AnyFunction]:
        """Map functions to a cell.

        Args:
            fs: functions
            map: Map from the reference cell to a physical cell
            inverse_map: Map to the reference cell from a physical cell
            tdim: The topological dimension of the cell

        Returns:
            Mapped functions
        """
        if self.mapping is not None:
            return super().perform_mapping(fs, map, inverse_map, tdim)
        out = []
        J = sympy.Matrix([[map[i].diff(x[j]) for j in range(tdim)] for i in range(tdim)])
        for dofs in zip(*[fs[i::tdim] for i in range(tdim)]):
            for i in range(tdim):
                f = ScalarFunction(0)
                for a, b in zip(dofs, J.row(i)):
                    f += a * b
                out.append(f)

        out2: typing.List[AnyFunction] = []
        for b in out:
            item = b.subs(x, inverse_map)
            out2.append(item)
        return out2

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        assert isinstance(self.point, VectorFunction)
        if len(self.point) == 1:
            desc = "v\\mapsto "
            desc += f"v'({','.join([_to_tex(i, True) for i in self.point])})"
            return desc, []
        desc = "v\\mapsto"
        desc += "\\frac{\\partial"
        if sum(self.derivative) > 1:
            desc += f"^{{{sum(self.derivative)}}}"
        desc += "}{"
        for v, i in zip("xyz", self.derivative):
            if i > 0:
                desc += f"\\partial {v}"
                if i > 1:
                    desc += f"^{{{i}}}"
        desc += "}"
        desc += f"v({','.join([_to_tex(i, True) for i in self.point])})"
        return desc, []

    name = "Point derivative evaluation"


class PointDirectionalDerivativeEvaluation(BaseFunctional):
    """A point evaluation of a derivative in a fixed direction."""

    def __init__(self, reference: Reference, point_in: FunctionInput, direction_in: FunctionInput,
                 entity: typing.Tuple[int, int], mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The points
            direction_in: The diection
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector
        self.dir = parse_function_input(direction_in)

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        d = self.dir.as_sympy()
        assert isinstance(d, tuple)
        return function.directional_derivative(d).subs(x, self.dof_point())

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        d = self.dir.as_sympy()
        assert isinstance(d, tuple)
        return d

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        assert isinstance(self.point, VectorFunction)
        assert isinstance(self.dir, VectorFunction)
        if len(self.point) == 1:
            desc = "v\\mapsto "
            desc += f"v'({','.join([_to_tex(i, True) for i in self.point])})"
            return desc, []
        desc = "v\\mapsto"
        desc += f"\\nabla{{v}}({','.join([_to_tex(i, True) for i in self.point])})"
        desc += "\\cdot\\left(\\begin{array}{c}"
        desc += "\\\\".join([_to_tex(i) for i in self.dir])
        desc += "\\end{array}\\right)"
        return desc, []

    name = "Point evaluation of directional derivative"


class PointNormalDerivativeEvaluation(PointDirectionalDerivativeEvaluation):
    """A point evaluation of a normal derivative."""

    def __init__(self, reference: Reference, point_in: FunctionInput, edge: Reference,
                 entity: typing.Tuple[int, int], mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The point
            edge: The edge of the cell
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        assert isinstance(edge, Interval)
        super().__init__(reference, point_in, edge.normal(), entity=entity, mapping=mapping)
        self.reference = edge

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        desc = "v\\mapsto"
        desc += "\\nabla{v}(" + ",".join([_to_tex(i, True) for i in self.dof_point()]) + ")"
        entity_n = self.entity_number()
        desc += "\\cdot\\hat{\\boldsymbol{n}}" + f"_{{{entity_n}}}"
        return desc, [
            "\\(\\hat{\\boldsymbol{n}}" + f"_{{{entity_n}}}\\) is the normal to facet {entity_n}"
        ]

    name = "Point evaluation of normal derivative"


class PointComponentSecondDerivativeEvaluation(BaseFunctional):
    """A point evaluation of a component of a second derivative."""

    def __init__(self, reference: Reference, point_in: FunctionInput,
                 component: typing.Tuple[int, int],
                 entity: typing.Tuple[int, int], mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The point
            component: The compenent of the derivative
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector
        self.component = component

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        return function.jacobian_component(self.component).subs(x, self.point)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        desc = "v\\mapsto"
        desc += "\\frac{\\partial^2v}{"
        for c in self.component:
            desc += "\\partial " + "xyz"[c]
        desc += "}(" + ",".join([_to_tex(i, True) for i in self.dof_point()]) + ")"
        return desc, []

    name = "Point evaluation of Jacobian component"


class PointInnerProduct(BaseFunctional):
    """An evaluation of an inner product at a point."""

    def __init__(self, reference: Reference, point_in: FunctionInput, lvec: FunctionInput,
                 rvec: FunctionInput,
                 entity: typing.Tuple[int, int], mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The points
            lvec: The vector to multiply on the left
            rvec: The vector to multiply on the right
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector
        self.lvec = parse_function_input(lvec)
        self.rvec = parse_function_input(rvec)
        assert self.lvec.is_vector
        assert self.rvec.is_vector

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        v = function.subs(x, self.point)

        return self.lvec.dot(v @ self.rvec)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        if self.rvec != self.lvec:
            return None
        lv = self.lvec.as_sympy()
        assert isinstance(lv, tuple)
        return lv

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        assert isinstance(self.lvec, VectorFunction)
        assert isinstance(self.rvec, VectorFunction)
        desc = "\\mathbf{V}\\mapsto"
        desc += "\\left(\\begin{array}{c}"
        desc += "\\\\".join([_to_tex(i) for i in self.lvec])
        desc += "\\end{array}\\right)^{\\text{t}}"
        desc += "\\mathbf{V}(" + ",".join([_to_tex(i, True) for i in self.dof_point()]) + ")"
        desc += "\\left(\\begin{array}{c}"
        desc += "\\\\".join([_to_tex(i) for i in self.rvec])
        desc += "\\end{array}\\right)"
        return desc, []

    name = "Point inner product"


class DotPointEvaluation(BaseFunctional):
    """A point evaluation in a given direction."""

    def __init__(self, reference: Reference, point_in: FunctionInput, vector_in: FunctionInput,
                 entity: typing.Tuple[int, int], mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The points
            vector_in: The vector to dot with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector
        self.vector = parse_function_input(vector_in)

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        v1 = function.subs(x, self.point)
        v2 = self.vector.subs(x, self.point)
        return v1.dot(v2)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        v = self.vector.as_sympy()
        if isinstance(v, tuple):
            return v
        return None

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        desc = "\\boldsymbol{v}\\mapsto"
        desc += "\\boldsymbol{v}(" + ",".join([_to_tex(i, True) for i in self.dof_point()]) + ")"
        if isinstance(self.vector, (tuple, list)):
            desc += "\\cdot\\left(\\begin{array}{c}"
            desc += "\\\\".join([_to_tex(i) for i in self.vector])
            desc += "\\end{array}\\right)"
        elif self.vector != 1:
            desc += f"\\cdot{_to_tex(self.vector)}"
        return desc, []

    name = "Dot point evaluation"


class IntegralAgainst(BaseFunctional):
    """An integral against a function."""

    f: AnyFunction

    def __init__(self, reference: Reference, integral_domain: Reference, f_in: FunctionInput,
                 entity: typing.Tuple[int, int], mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            f_in: The function to multiply with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.integral_domain = integral_domain

        f = parse_function_input(f_in)
        f = f.subs(x, t)

        if f.is_vector:
            assert len(f) == self.integral_domain.tdim
            self.f = mappings.contravariant(
                f, integral_domain.get_map_to_self(), integral_domain.get_inverse_map_to_self(),
                integral_domain.tdim)
        elif f.is_matrix:
            assert f.shape[0] == self.integral_domain.tdim
            assert f.shape[1] == self.integral_domain.tdim
            self.f = mappings.double_contravariant(
                f, integral_domain.get_map_to_self(), integral_domain.get_inverse_map_to_self(),
                integral_domain.tdim)
        else:
            self.f = f

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        return tuple(sympy.Rational(sum(i), len(i)) for i in zip(*self.integral_domain.vertices))

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        point = [i for i in self.integral_domain.origin]
        for i, a in enumerate(zip(*self.integral_domain.axes)):
            for j, k in zip(a, t):
                point[i] += j * k
        v1 = function.subs(x, point)
        integrand = self.dot(v1)
        return integrand.integral(self.integral_domain)

    def dot(self, function: AnyFunction) -> ScalarFunction:
        """Dot a function with the moment function.

        Args:
            function: The function

        Returns:
            The product of the function and the moment function
        """
        return function.dot(self.f)

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        if isinstance(self.f, tuple):
            desc = "\\mathbf{v}\\mapsto"
            desc += f"\\displaystyle\\int_{{{entity}}}"
            desc += _to_tex(self.f, True) + "\\cdot"
            desc += "\\mathbf{v}"
        else:
            desc = "\\mathbf{v}\\mapsto"
            desc += f"\\displaystyle\\int_{{{entity}}}"
            if self.f != 1:
                desc += "(" + _to_tex(self.f, True) + ")"
            desc += "v"
        return desc, [entity_def]

    name = "Integral against"


class IntegralOfDivergenceAgainst(BaseFunctional):
    """An integral of the divergence against a function."""

    def __init__(self, reference: Reference, integral_domain: Reference, f_in: FunctionInput,
                 entity: typing.Tuple[int, int], mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            f_in: The function to multiply with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.integral_domain = integral_domain

        f = parse_function_input(f_in)
        self.f = f.subs(x, t)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        return tuple(sympy.Rational(sum(i), len(i)) for i in zip(*self.integral_domain.vertices))

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        point = [i for i in self.integral_domain.origin]
        for i, a in enumerate(zip(*self.integral_domain.axes)):
            for j, k in zip(a, t):
                point[i] += j * k
        v1 = function.div().subs(x, point)
        integrand = self.dot(v1)
        return integrand.integral(self.integral_domain)

    def dot(self, function: ScalarFunction) -> ScalarFunction:
        """Dot a function with the moment function.

        Args:
            function: The function

        Returns:
            The product of the function and the moment function
        """
        return function * self.f

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        desc = "\\mathbf{v}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        if self.f != 1:
            desc += "(" + _to_tex(self.f, True) + ")"
        desc += "\\nabla\\cdot\\mathbf{v}"
        return desc, [entity_def]

    name = "Integral of divergence against"


class IntegralOfDirectionalMultiderivative(BaseFunctional):
    """An integral of a directional derivative of a scalar function."""

    def __init__(self, reference: Reference, integral_domain: Reference, directions: SetOfPoints,
                 orders: typing.Tuple[int, ...], entity: typing.Tuple[int, int], scale: int = 1,
                 mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            directions: The diections of the derivatives
            orders: The orders of the derivatives
            entity: The entity the functional is associated with
            scale: The scale factor
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.integral_domain = integral_domain
        self.directions = directions
        self.orders = orders
        self.scale = scale

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        return tuple(sympy.Rational(sum(i), len(i)) for i in zip(*self.integral_domain.vertices))

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        for dir, o in zip(self.directions, self.orders):
            for i in range(o):
                function = function.grad(len(dir)).dot(VectorFunction(dir))
        point = [i for i in self.integral_domain.origin]
        for i, a in enumerate(zip(*self.integral_domain.axes)):
            for j, k in zip(a, t):
                point[i] += j * k
        integrand = ScalarFunction(self.scale) * function.subs(x, point)
        return integrand.integral(self.integral_domain)

    def perform_mapping(
        self, fs: typing.List[AnyFunction], map: PointType, inverse_map: PointType, tdim: int
    ) -> typing.List[AnyFunction]:
        """Map functions to a cell.

        Args:
            fs: functions
            map: Map from the reference cell to a physical cell
            inverse_map: Map to the reference cell from a physical cell
            tdim: The topological dimension of the cell

        Returns:
            Mapped functions
        """
        if sum(self.orders) > 0:
            raise NotImplementedError("Mapping high order derivatives not implemented")
        return super().perform_mapping(fs, map, inverse_map, tdim)

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        desc = "\\mathbf{V}\\mapsto"
        desc += "\\displaystyle"
        if self.scale != 1:
            desc += _to_tex(self.scale)
        desc += f"\\int_{{{entity}}}"
        for order, dir in zip(self.orders, self.directions):
            if order > 0:
                desc += "\\frac{\\partial"
                if order > 1:
                    desc += f"^{{{order}}}"
                desc += "}{"
                desc += "\\partial" + _to_tex(dir)
                if order > 1:
                    desc += f"^{{{order}}}"
                desc += "}"
        desc += "v"
        return desc, [entity_def]

    name = "Integral of a directional derivative"


class IntegralMoment(BaseFunctional):
    """An integral moment."""

    f: AnyFunction

    def __init__(self, reference: Reference, integral_domain: Reference,
                 f_in: FunctionInput, dof: BaseFunctional,
                 entity: typing.Tuple[int, int], mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.integral_domain = integral_domain
        self.dof = dof

        f = parse_function_input(f_in)
        f = f.subs(x, t)

        if f.is_vector:
            assert len(f) == self.integral_domain.tdim
            self.f = mappings.contravariant(
                f, integral_domain.get_map_to_self(), integral_domain.get_inverse_map_to_self(),
                integral_domain.tdim)
        elif f.is_matrix:
            assert f.shape[0] == self.integral_domain.tdim
            assert f.shape[1] == self.integral_domain.tdim
            self.f = mappings.double_contravariant(
                f, integral_domain.get_map_to_self(), integral_domain.get_inverse_map_to_self(),
                integral_domain.tdim)
        else:
            self.f = f

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        point = [i for i in self.integral_domain.origin]
        for i, a in enumerate(zip(*self.integral_domain.axes)):
            for j, k in zip(a, t):
                point[i] += j * k

        v1 = function.subs(x, point)
        integrand = self.dot(v1)
        return integrand.integral(self.integral_domain)

    def dot(self, function: AnyFunction) -> ScalarFunction:
        """Dot a function with the moment function.

        Args:
            function: The function

        Returns:
            The product of the function and the moment function
        """
        return function.dot(self.f)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        p = self.dof.dof_point()
        return tuple(
            o + sum(self.integral_domain.axes[j][i] * c for j, c in enumerate(p))
            for i, o in enumerate(self.integral_domain.origin)
        )

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        p = self.dof.dof_direction()
        if p is None:
            return None
        vp = VectorFunction(p)
        out = []
        for i in range(self.integral_domain.gdim):
            entry = vp.dot(VectorFunction([a[i] for a in self.integral_domain.axes])).as_sympy()
            assert isinstance(entry, sympy.core.expr.Expr)
            out.append(entry)
        return tuple(out)

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        if isinstance(self.f, tuple):
            if len(self.f) in [self.reference.tdim, self.integral_domain.tdim]:
                desc = "\\boldsymbol{v}\\mapsto"
                desc += f"\\displaystyle\\int_{{{entity}}}"
                desc += "\\boldsymbol{v}\\cdot"
                desc += "\\left(\\begin{array}{c}"
                desc += "\\\\".join([_to_tex(i) for i in self.f])
                desc += "\\end{array}\\right)"
            else:
                if len(self.f) == self.integral_domain.tdim ** 2:
                    size = self.integral_domain.tdim
                elif len(self.f) == self.reference.tdim ** 2:
                    size = self.reference.tdim
                else:
                    raise NotImplementedError()
                desc = "\\mathbf{V}\\mapsto"
                desc += f"\\displaystyle\\int_{{{entity}}}"
                desc += "\\mathbf{V}:"
                desc += "\\left(\\begin{array}{" + "c" * size + "}"
                desc += "\\\\".join(["&".join(
                    [_to_tex(self.f[i]) for i in range(size * row, size * (row + 1))]
                ) for row in range(size)])
                desc += "\\end{array}\\right)"
        else:
            desc = "v\\mapsto"
            desc += f"\\displaystyle\\int_{{{entity}}}"
            if self.f != 1:
                desc += "(" + _to_tex(self.f) + ")"
            desc += "v"
        return desc, [entity_def]

    name = "Integral moment"


class VecIntegralMoment(IntegralMoment):
    """An integral moment applied to a component of a vector."""

    def __init__(self, reference: Reference, integral_domain: Reference, f_in: FunctionInput,
                 dot_with_in: FunctionInput, dof: BaseFunctional, entity: typing.Tuple[int, int],
                 mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            f_in: The function to multiply with
            dot_with_in: The vector to take the dot product with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        f = parse_function_input(f_in)
        super().__init__(reference, integral_domain, f, dof, entity=entity, mapping=mapping)
        self.dot_with = parse_function_input(dot_with_in)

    def dot(self, function: AnyFunction) -> ScalarFunction:
        """Dot a function with the moment function.

        Args:
            function: The function

        Returns:
            The product of the function and the moment function
        """
        return function.dot(self.dot_with) * self.f

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        dw = self.dot_with.as_sympy()
        if isinstance(dw, tuple):
            return dw
        return None

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        desc = "\\boldsymbol{v}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        if self.f != 1:
            desc += "(" + _to_tex(self.f, True) + ")"
        desc += "\\boldsymbol{v}\\cdot"
        desc += _to_tex(self.dot_with)
        return desc, [entity_def]

    name = "Vector integral moment"


class DerivativeIntegralMoment(IntegralMoment):
    """An integral moment of the derivative of a scalar function."""

    def __init__(self, reference: Reference, integral_domain: Reference, f: FunctionInput,
                 dot_with_in: FunctionInput, dof: BaseFunctional, entity: typing.Tuple[int, int],
                 mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            f: The function to multiply with
            dot_with_in: The vector to take the dot product with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, integral_domain, f, dof, entity=entity, mapping=mapping)
        self.dot_with = parse_function_input(dot_with_in)

    def dot(self, function: AnyFunction) -> ScalarFunction:
        """Dot a function with the moment function.

        Args:
            function: The function

        Returns:
            The product of the function and the moment function
        """
        assert function.is_vector
        assert self.f.is_scalar
        return function.dot(self.dot_with) * self.f

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        dw = self.dot_with.as_sympy()
        assert isinstance(dw, tuple)
        return dw

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        point = [i for i in self.integral_domain.origin]
        for i, a in enumerate(zip(*self.integral_domain.axes)):
            for j, k in zip(a, t):
                point[i] += j * k
        integrand = self.dot(function.grad(self.integral_domain.gdim).subs(x, point))
        value = integrand.integral(self.integral_domain)
        return value

    name = "Derivative integral moment"


class DivergenceIntegralMoment(IntegralMoment):
    """An integral moment of the divergence of a vector function."""

    def __init__(self, reference: Reference, integral_domain: Reference, f_in: FunctionInput,
                 dof: BaseFunctional, entity: typing.Tuple[int, int],
                 mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        f = parse_function_input(f_in)
        assert f.is_scalar
        super().__init__(reference, integral_domain, f, dof, entity=entity, mapping=mapping)

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        point = [i for i in self.integral_domain.origin]
        for i, a in enumerate(zip(*self.integral_domain.axes)):
            for j, k in zip(a, t):
                point[i] += j * k
        assert function.is_vector
        integrand = self.dot(function.div().subs(x, point))
        return integrand.integral(self.integral_domain)

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        desc = "\\boldsymbol{v}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        if self.f != 1:
            desc += "(" + _to_tex(self.f, True) + ")"
        desc += "\\nabla\\cdot\\boldsymbol{v}"
        return desc, [entity_def]

    name = "Integral moment of divergence"


class TangentIntegralMoment(VecIntegralMoment):
    """An integral moment in the tangential direction."""

    def __init__(self, reference: Reference, integral_domain: Reference, f_in: FunctionInput,
                 dof: BaseFunctional, entity: typing.Tuple[int, int],
                 mapping: typing.Union[str, None] = "covariant"):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        f = parse_function_input(f_in)
        assert f.is_scalar
        super().__init__(reference, integral_domain, f, integral_domain.tangent(), dof,
                         entity=entity, mapping=mapping)

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_n = self.entity_number()
        entity_def = self.entity_definition()
        desc = "\\boldsymbol{v}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        desc += "\\boldsymbol{v}\\cdot"
        if self.f != 1:
            desc += "(" + _to_tex(self.f) + ")"
        desc += "\\hat{\\boldsymbol{t}}" + f"_{{{entity_n}}}"
        return desc, [
            entity_def,
            f"\\(\\hat{{\\boldsymbol{{t}}}}_{{{entity_n}}}\\) is the tangent to edge {entity_n}"
        ]

    name = "Tangential integral moment"


class NormalIntegralMoment(VecIntegralMoment):
    """An integral moment in the normal direction."""

    def __init__(self, reference: Reference, integral_domain: Reference, f_in: FunctionInput,
                 dof: BaseFunctional, entity: typing.Tuple[int, int],
                 mapping: typing.Union[str, None] = "contravariant"):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        f = parse_function_input(f_in)
        assert f.is_scalar
        super().__init__(reference, integral_domain, f, integral_domain.normal(), dof,
                         entity=entity, mapping=mapping)

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_n = self.entity_number()
        entity_def = self.entity_definition()
        desc = "\\boldsymbol{v}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        desc += "\\boldsymbol{v}\\cdot"
        if self.f != 1:
            desc += "(" + _to_tex(self.f, True) + ")"
        desc += "\\hat{\\boldsymbol{n}}" + f"_{{{entity_n}}}"
        return desc, [
            entity_def,
            f"\\(\\hat{{\\boldsymbol{{n}}}}_{{{entity_n}}}\\) is the normal to facet {entity_n}"
        ]

    name = "Normal integral moment"


class NormalDerivativeIntegralMoment(DerivativeIntegralMoment):
    """An integral moment in the normal direction."""

    def __init__(self, reference: Reference, integral_domain: Reference, f_in: FunctionInput,
                 dof: BaseFunctional, entity: typing.Tuple[int, int],
                 mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        f = parse_function_input(f_in)
        assert f.is_scalar
        super().__init__(reference, integral_domain, f, integral_domain.normal(), dof,
                         entity=entity, mapping=mapping)

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_n = self.entity_number()
        entity_def = self.entity_definition()
        desc = "v\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        if self.f != 1:
            desc += "(" + _to_tex(self.f, True) + ")"
        desc += "\\frac{\\partial v}"
        desc += "{\\partial\\hat{\\boldsymbol{n}}" + f"_{{{entity_n}}}" + "}"
        return desc, [
            entity_def,
            f"\\(\\hat{{\\boldsymbol{{n}}}}_{{{entity_n}}}\\) is the normal to facet {entity_n}"
        ]

    name = "Normal derivative integral moment"


class InnerProductIntegralMoment(IntegralMoment):
    """An integral moment of the inner product with a vector."""

    def __init__(self, reference: Reference, integral_domain: Reference, f_in: FunctionInput,
                 inner_with_left_in: FunctionInput, inner_with_right_in: FunctionInput,
                 dof: BaseFunctional, entity: typing.Tuple[int, int],
                 mapping: typing.Union[str, None] = "identity"):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            inner_with_left_in: The vector to multiply on the left
            inner_with_right_in: The vector to multiply on the right
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        f = parse_function_input(f_in)
        inner_with_left = parse_function_input(inner_with_left_in)
        inner_with_right = parse_function_input(inner_with_right_in)
        assert f.is_scalar
        assert inner_with_left.is_vector
        assert inner_with_right.is_vector

        super().__init__(reference, integral_domain, f, dof, entity=entity, mapping=mapping)
        self.inner_with_left = inner_with_left
        self.inner_with_right = inner_with_right

    def dot(self, function: AnyFunction) -> ScalarFunction:
        """Take the inner product of a function with the moment direction.

        Args:
            function: The function

        Returns:
            The inner product of the function and the moment direction
        """
        assert function.is_matrix
        return self.inner_with_left.dot(
            function @ self.inner_with_right) * self.f * self.integral_domain.jacobian()

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        if self.inner_with_left != self.inner_with_right:
            return None
        il = self.inner_with_left.as_sympy()
        assert isinstance(il, tuple)
        return il

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        desc = "\\boldsymbol{V}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        if self.f != 1:
            desc += "(" + _to_tex(self.f, True) + ")"
        desc += _to_tex(self.inner_with_left) + "^{\\text{t}}"
        desc += "\\boldsymbol{V}"
        desc += _to_tex(self.inner_with_right)
        return desc, [entity_def]

    name = "Inner product integral moment"


class NormalInnerProductIntegralMoment(InnerProductIntegralMoment):
    """An integral moment of the inner product with the normal direction."""

    def __init__(self, reference: Reference, integral_domain: Reference, f_in: FunctionInput,
                 dof: BaseFunctional, entity: typing.Tuple[int, int],
                 mapping: typing.Union[str, None] = "double_contravariant"):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        f = parse_function_input(f_in)
        assert f.is_scalar
        super().__init__(reference, integral_domain, f, integral_domain.normal(),
                         integral_domain.normal(), dof, entity=entity, mapping=mapping)

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_n = self.entity_number()
        entity_def = self.entity_definition()
        desc = "\\mathbf{V}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        if self.f != 1:
            desc += "(" + _to_tex(self.f, True) + ")"
        desc += f"|{{{entity}}}|"
        desc += "\\hat{\\boldsymbol{n}}^{\\text{t}}" + f"_{{{entity_n}}}"
        desc += "\\mathbf{V}"
        desc += "\\hat{\\boldsymbol{n}}" + f"_{{{entity_n}}}"
        return desc, [
            entity_def,
            f"\\(\\hat{{\\boldsymbol{{n}}}}_{{{entity_n}}}\\) is the normal to facet {entity_n}"
        ]

    name = "Normal inner product integral moment"


# Types
ListOfFunctionals = typing.List[BaseFunctional]



"""Hermite elements on simplices.

This element's definition appears in https://doi.org/10.1016/0045-7825(72)90006-0
(Ciarlet, Raviart, 1972)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import DerivativePointEvaluation, ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..polynomials import polynomial_set_1d
from ..references import Reference


class Hermite(CiarletElement):
    """Hermite finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        assert order == 3
        dofs: ListOfFunctionals = []
        for v_n, v in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
            for i in range(reference.tdim):
                dofs.append(DerivativePointEvaluation(
                    reference, v, tuple(1 if i == j else 0 for j in range(reference.tdim)),
                    entity=(0, v_n)))
        for e_n in range(reference.sub_entity_count(2)):
            sub_entity = reference.sub_entity(2, e_n)
            dofs.append(PointEvaluation(reference, sub_entity.midpoint(), entity=(2, e_n)))

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    names = ["Hermite"]
    references = ["interval", "triangle", "tetrahedron"]
    min_order = 3
    max_order = 3
    continuity = "C1"



"""Hsieh-Clough-Tocher elements on simplices.

This element's definition appears in https://doi.org/10.2307/2006147
(Ciarlet, 1978)
"""

import typing

import sympy

from ..basis_functions import BasisFunction
from ..finite_element import CiarletElement
from ..functionals import (DerivativePointEvaluation, ListOfFunctionals, PointEvaluation,
                           PointNormalDerivativeEvaluation)
from ..functions import FunctionInput, VectorFunction
from ..piecewise_functions import PiecewiseFunction
from ..references import Reference
from .hermite import Hermite


class HsiehCloughTocher(CiarletElement):
    """Hsieh-Clough-Tocher finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        from symfem import create_reference
        assert order == 3
        assert reference.name == "triangle"
        dofs: ListOfFunctionals = []
        for v_n, vs in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, vs, entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, vs, (1, 0), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, vs, (0, 1), entity=(0, v_n)))
        for e_n in range(reference.sub_entity_count(1)):
            sub_ref = reference.sub_entity(1, e_n)
            dofs.append(PointNormalDerivativeEvaluation(
                reference, sub_ref.midpoint(), sub_ref, entity=(1, e_n)))

        mid = tuple(sympy.Rational(sum(i), len(i)) for i in zip(*reference.vertices))

        subs = [
            (reference.vertices[0], reference.vertices[1], mid),
            (reference.vertices[1], reference.vertices[2], mid),
            (reference.vertices[2], reference.vertices[0], mid)]

        refs = [create_reference("triangle", vs) for vs in subs]

        hermite_spaces = [Hermite(ref, 3) for ref in refs]

        piece_list: typing.List[typing.Tuple[typing.Union[int, BasisFunction], ...]] = []
        piece_list.append((hermite_spaces[0].get_basis_function(0), 0,
                           hermite_spaces[2].get_basis_function(3)))
        piece_list.append((hermite_spaces[0].get_basis_function(1), 0,
                           hermite_spaces[2].get_basis_function(4)))
        piece_list.append((hermite_spaces[0].get_basis_function(2), 0,
                           hermite_spaces[2].get_basis_function(5)))
        piece_list.append((hermite_spaces[0].get_basis_function(3),
                           hermite_spaces[1].get_basis_function(0), 0))
        piece_list.append((hermite_spaces[0].get_basis_function(4),
                           hermite_spaces[1].get_basis_function(1), 0))
        piece_list.append((hermite_spaces[0].get_basis_function(5),
                           hermite_spaces[1].get_basis_function(2), 0))
        piece_list.append((hermite_spaces[0].get_basis_function(6),
                           hermite_spaces[1].get_basis_function(6),
                           hermite_spaces[2].get_basis_function(6)))
        piece_list.append((hermite_spaces[0].get_basis_function(7),
                           hermite_spaces[1].get_basis_function(7),
                           hermite_spaces[2].get_basis_function(7)))
        piece_list.append((hermite_spaces[0].get_basis_function(8),
                           hermite_spaces[1].get_basis_function(8),
                           hermite_spaces[2].get_basis_function(8)))
        piece_list.append((0, hermite_spaces[1].get_basis_function(3),
                           hermite_spaces[2].get_basis_function(0)))
        piece_list.append((0, hermite_spaces[1].get_basis_function(4),
                           hermite_spaces[2].get_basis_function(1)))
        piece_list.append((0, hermite_spaces[1].get_basis_function(5),
                           hermite_spaces[2].get_basis_function(2)))

        # TODO: are these right to remove??
        # piece_list.append((hermite_spaces[0].get_basis_function(9), 0, 0))
        # piece_list.append((0, hermite_spaces[1].get_basis_function(9), 0))
        # piece_list.append((0, 0, hermite_spaces[2].get_basis_function(9)))

        piece_list2: typing.List[VectorFunction] = []
        for i in piece_list:
            piece_list2.append(VectorFunction(i))

        poly: typing.List[FunctionInput] = []
        poly += [
            PiecewiseFunction({i: j for i, j in zip(subs, p)}, 2)
            for p in piece_list2]

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )

    names = ["Hsieh-Clough-Tocher", "Clough-Tocher", "HCT", "CT"]
    references = ["triangle"]
    min_order = 3
    max_order = 3
    continuity = "C1"
